"""Provides the :class:`ParametricCurve` class."""
# pylint: disable=invalid-name,too-many-arguments

from __future__ import annotations
from typing import Callable, Optional, Tuple, Union
import functools

import scipy.integrate as integrate
import scipy.optimize as optimize

import numpy as np

import sympy
import sympy.geometry as sympy_geom

import splipy

from fibomat.shapes.shape import Shape
from fibomat.shapes.arc_spline import ArcSplineCompatible, ArcSpline
from fibomat.linalg import VectorLike, Vector, BoundingBox
from fibomat.curve_tools.biarc_approximation import approximate_parametric_curve


class ParametricCurve(Shape, ArcSplineCompatible):
    """Parametric curve  f: [a, b] -> R^2 with f \\in C^\\inf

    E.g. f(u) = (cos(u), sin(u)).

    TODO: handle bounding box!
    TODO: Put references here (thesis + 2 papers)
    """

    def __init__(
        self,
        func: Callable[[np.ndarray], np.ndarray],
        d_func: Callable[[np.ndarray], np.ndarray],
        d2_func: Callable[[np.ndarray], np.ndarray],
        domain: Tuple[float, float],
        bounding_box: BoundingBox,
        curvature: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        length: Optional[Callable[[float, float], float]] = None,
        description: Optional[str] = None
    ):
        r"""
        All `Callable`\ s must be vectorized. Hence

            - func(1) should return np.array([x, y])
            - func([1, 2]) should return np.array([[x1, y1], [x2, y2]])

        If `length` is not provided, it will be calculated numerically.

        Args:
            func (Callable[[np.ndarray], np.ndarray]): function value of parametric curve
            d_func (Callable[[np.ndarray], np.ndarray]): function value of first derivative
            d2_func (Callable[[np.ndarray], np.ndarray]): function value of second derivative
            domain (Tuple[float, float]): parametric domain of curve
            bounding_box (BoundingBox): bounding box of curve
            curvature (Optional[Callable[[np.ndarray], np.ndarray]], optional): curvature of parametric curve.
            length (Optional[Callable[[float, float], float]], optional):
                arc length of parametric curve in interval u_1, u_2
            description (str, optional): description
        """
        super().__init__(description)

        self._func = func
        self._d_func = d_func
        self._d2_func = d2_func

        if curvature:
            self._curvature = curvature
        else:
            def curvature_impl(t: np.ndarray):
                df = self._d_func(t)
                d2f = self._d2_func(t)
                # print(t, type(t), t.ndim)
                if t.ndim == 0:
                    return (df[0] * d2f[1] - df[1] * d2f[0]) / (df[0]**2 + df[1]**2)**1.5

                return (df[:, 0] * d2f[:, 1] - df[:, 1] * d2f[:, 0]) / (df[:, 0]**2 + df[:, 1]**2)**1.5

            self._curvature = curvature_impl

        if length:
            self._length = length
        else:
            def length_impl(t_0, t_1):
                return integrate.quad(lambda t: np.linalg.norm(self._d_func(t)), t_0, t_1)[0]

            self._length = length_impl

        self._domain = domain

    @classmethod
    def from_sympy_curve(
        cls,
        curve: sympy_geom.Curve,
        try_length_integration: bool = False,
        description: Optional[str] = None
    ):
        """Create a :class:`ParametricCurve` from a sympy curve. All derivatives are calculated automatically.

        Args:
            curve (sympy.geometry.Curve): parametric curve.
            try_length_integration (bool):
                if True, it is attempted to solve arc length parametrization integral analytically.
            description (str, optional): description

        Returns:
            ParametricCurve
        """
        # pylint: disable=function-redefined

        # https://github.com/sympy/sympy/issues/5642
        # https://stackoverflow.com/a/59757810
        def lambdify_np(arg, exp, modules=None):
            exp_lambdified = sympy.lambdify(arg, exp, modules)
            if exp.is_constant:
                return lambda t: np.full_like(t, exp_lambdified(t))

            return exp_lambdified

        def lambdify_np_2d(arg, expressions, modules=None):
            l1 = lambdify_np(arg, expressions[0], modules)
            l2 = lambdify_np(arg, expressions[1], modules)
            # return lambda t: np.vstack((l1(t), l2(t))).T
            return lambda t: np.squeeze(np.vstack((l1(t), l2(t))).T)
            # return lambda t: np.c_[l1(t), l2(t)]

        # TODO: check if diff failed !
        dfunc = [sympy.diff(curve.functions[0], curve.parameter), sympy.diff(curve.functions[1], curve.parameter)]
        d2func = [sympy.diff(dfunc[0], curve.parameter), sympy.diff(dfunc[1], curve.parameter)]

        curvature = sympy.simplify(
            (dfunc[0]*d2func[1] - dfunc[1]*d2func[0]) / (dfunc[0]**2 + dfunc[1]**2)**sympy.Rational(3, 2)
        )

        # todo: time out integration
        lambda_length: Optional[Callable]
        if try_length_integration:
            length = sympy.simplify(sympy.integrate(sympy.sqrt(dfunc[0]*dfunc[0] + dfunc[1]*dfunc[1]), curve.parameter))

            # integration failed
            if isinstance(length, sympy.Integral):
                lambda_length = None
            else:
                # lambda_length = sympy.lambdify(curve.parameter, length, 'numpy')
                antiderivate = lambdify_np(curve.parameter, length, 'numpy')

                def lambda_length(t_0: float, t_1: float):
                    return antiderivate(t_1) - antiderivate(t_0)
        else:
            lambda_length = None

        return cls(
            func=lambdify_np_2d(curve.parameter, curve.functions, 'numpy'),
            d_func=lambdify_np_2d(curve.parameter, dfunc, 'numpy'),
            d2_func=lambdify_np_2d(curve.parameter, d2func, 'numpy'),
            domain=(float(curve.limits[1]), float(curve.limits[2])),
            bounding_box=BoundingBox((0, 0), (0, 0)),
            curvature=lambdify_np(curve.parameter, curvature, 'numpy'),
            length=lambda_length,
            description=description
        )

    @classmethod
    def from_splipy_curve(cls, splipy_curve: splipy.Curve, description: Optional[str] = None):
        splipy_curve = splipy_curve.clone()

        func = lambda t: splipy_curve.evaluate(t)
        d_func = lambda t: splipy_curve.derivative(t, d=1)
        d2_func = lambda t: splipy_curve.derivative(t, d=2)
        curvature = lambda t: splipy_curve.curvature(t)
        length = lambda t_0, t_1: splipy_curve.length(t_0, t_1)
        domain = splipy_curve.start(direction=0), splipy_curve.end(direction=0)

        print(domain, type(domain[0]))

        return cls(
            func=func,
            d_func=d_func,
            d2_func=d2_func,
            domain=domain,
            bounding_box=BoundingBox((0, 0), (0, 0)),
            curvature=curvature,
            length=length,
            description=description
        )

    def __repr__(self) -> str:
        return '{}(...)'.format(self.__class__.__name__)

    def to_arc_spline(self, rasterize_pitch: Optional[float] = None, epsilon: Optional[float] = None) -> ArcSpline:
        if (rasterize_pitch is not None and epsilon is None) or (rasterize_pitch is None and epsilon is not None):
            raise ValueError('Either rasterize_pitch and epsilon must be given or none of them.')
        # if (rasterize_pitch is not None and epsilon is not None) or (rasterize_pitch is None and epsilon is None):
        #     raise ValueError('Either rasterize_pitch and epsilon must be given or none of them.')

        if rasterize_pitch is None:
            rasterize_pitch = self.length / 100
            epsilon = rasterize_pitch / 10
        

        return ArcSpline.from_segments(approximate_parametric_curve(self, rasterize_pitch, epsilon))

    @property
    def domain(self) -> Tuple[float, float]:
        """Parametric domain of curve.

        Access:
            get

        Returns:
            Tuple[float, float]
        """
        return self._domain

    @property
    def is_closed(self) -> bool:
        return np.allclose(self._func(self._domain[0]), self._func(self._domain[1]))

    @property
    def length(self) -> float:
        """Arc length of curve.

        Access:
            get

        Returns:
            float
        """
        return self._length(*self._domain)

    def f(self, t: Union[float, np.ndarray]) -> np.ndarray:
        """Function values of param. curve.

        Args:
            t (float, np.ndarray): time points for evaluation.

        Returns:
            np.ndarray
        """
        return self._func(np.asarray(t))

    def df(self, t):
        """Function values of first derivative of param. curve.

        Args:
            t (float, np.ndarray): time points for evaluation.

        Returns:
            np.ndarray
        """
        return self._d_func(np.asarray(t))

    def d2f(self, t):
        """Function values of second derivative of param. curve.

        Args:
            t (float, np.ndarray): time points for evaluation.

        Returns:
            np.ndarray
        """
        return self._d2_func(np.asarray(t))

    def curvature(self, t):
        """Curvature of param. curve.

        Args:
            t (float, np.ndarray): time points for evaluation.

        Returns:
            np.ndarray
        """
        return self._curvature(np.asarray(t))

    def rasterize(
        self,
        pitch: float,
        domain: Optional[Tuple[float, float]] = None,
        safety: float = 1.25,
        add_endpoint: bool = False
    ) -> np.ndarray:
        """Rasterize the param. curve equally.

        Args:
            pitch (float): distance of rasteruized points on the curve.
            domain (Tuple[float, float], optional): parametric domain to be used. Default to self.domain.
            safety (float):
                The upper bound of the function parameter of a point is estimated with the tangent vector `t` of the
                previous point with `t_next_up = t_prev + pitch * safety / ||t||`. If the function has large gradients,
                thesafety factor must be increased. Otherwise, `t_next_up` is not an upper bound anymore. Default to 1.5
            add_endpoint (bool): if True, the point at f(domain[1]) is added to the rasterized points, if the distance
                                 to the point before is smaller than the pitch.


        Returns:
            np.ndarray: function parameters (NOT function values)

        Raises:
            RuntimeError: Raised if `safety` is to small.
        """
        safety_step = .05
        safety = 1.

        if domain:
            n_points = int(self._length(*domain) / pitch) + 1
        else:
            n_points = int(self._length(*self._domain) / pitch) + 1

        if domain:
            t_min, t_max = domain
        else:
            t_min, t_max = self._domain

        t = t_min

        # create a little more space than needed to account for inaccuracies at the rasterization step
        buffer_size = int(n_points * 1.2)
        t_res = np.empty(buffer_size, dtype=float)
        t_res[0] = t_min

        def length_diff(t_0: float, t_1: float):
            return self._length(t_0, t_1) - pitch

        i = 1
        # for i in range(1, n_points):
        while t < t_max:
            # f = lambda t_: self._length(t, t_) - pitch
            root_found = False

            df = np.linalg.norm(self.df(t))
            if np.isclose(df, 0):
                df = (t_max - t_min) / n_points

            while not root_found:
                try:
                    t_bound = t + safety * pitch / df
                    if t_bound > t_max:
                        t_bound = t_max

                    # print(functools.partial(length_diff, t)(t), functools.partial(length_diff, t)(t_bound))

                    t = optimize.toms748(
                        functools.partial(length_diff, t),
                        t, t_bound,
                        xtol=0.01*pitch, rtol=0.001*pitch
                    )  # t_max
                except ValueError:
                    if t_bound == t_max:
                        # if self._length(*domain) is a multiple of the pitch: f(t) < 0 and f(t_max) \approx 0
                        # (hence toms748 would fail because no sign change occurs)
                        t = t_max
                        root_found = True
                    else:
                        # print('increase')
                        safety += safety_step
                else:
                    root_found = True
            t_res[i] = t
            i += 1

            if i > buffer_size:
                raise RuntimeError('i >= buffer_size. This is probably a bug. Please report it.')

        if add_endpoint and not np.isclose(t_res[i-2], t_max):
            t_res[i-1] = t_max
            i += 1

        t_res = np.resize(t_res, i-1)

        return t_res

    def rasterize_at(self, pitch: float, domain: Optional[Tuple[float, float]] = None):
        """Rasterize the param. curve equally.

        Args:
            pitch (float): distance of rasteruized points on the curve.
            domain (Tuple[float, float], optional): parametric domain to be used. Default to self.domain.

        Returns:
             np.ndarray: function values
        """
        return self.f(self.rasterize(pitch, domain))

    @property
    def bounding_box(self) -> BoundingBox:
        raise NotImplementedError(
            f'Cannot calculate a bounding box for {self.__class__.__name__}. '
            'Convert it to an ArcSpline first.'
        )

    @property
    def center(self) -> Vector:
        raise NotImplementedError(
            f'Cannot calculate center for {self.__class__.__name__}. '
            'Convert it to an ArcSpline first.'
        )

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        raise NotImplementedError(f'Cannot translate a {self.__class__.__name__}. Convert it to an ArcSpline first.')

    def _impl_rotate(self, theta: float) -> None:
        raise NotImplementedError(f'Cannot rotate a {self.__class__.__name__}. Convert it to an ArcSpline first.')

    def _impl_scale(self, fac: float) -> None:
        raise NotImplementedError(f'Cannot scale a {self.__class__.__name__}. Convert it to an ArcSpline first.')

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        raise RuntimeError(f'Cannot mirror a {self.__class__.__name__}. Convert it to an ArcSpline first.')



"""Provides the :class:`Rect` class."""
from __future__ import annotations
from typing import Optional, List

import numpy as np

from fibomat.shapes import Shape
from fibomat.shapes.arc_spline import ArcSplineCompatible, ArcSpline
from fibomat.linalg import VectorLike, Vector, BoundingBox, signed_angle_between
from fibomat.shapes.line import Line


class Rect(Shape, ArcSplineCompatible):
    """2-dim rect."""

    def __init__(  # pylint: disable=too-many-arguments
            self,
            width: float, height: float, theta: float = 0,
            center: Optional[VectorLike] = None,
            description: Optional[str] = None
    ):
        """
        Args:
            width (float): width of rect
            height (float): height of rect
            theta (float): rotation angle of rect, default to 0
            center (VectorLike, optional): center of rect, default to (0, 0)
            description (str, optional): description
        """
        super().__init__(description)

        width_half = float(width) / 2
        height_half = float(height) / 2

        center = Vector(center)

        self._rect = ArcSpline([
            (*(center + (width_half, height_half)), 0.),
            (*(center + (-width_half, height_half)), 0.),
            (*(center + (-width_half, -height_half)), 0.),
            (*(center + (width_half, -height_half)), 0.)
        ], True).rotated(theta, 'center')

        self._rot_angle = float(theta)

        # self._axes = VectorArray([float(width) / 2, 0.], [0., float(height) / 2])
        # self._axes = self._axes.rotated(theta)
        #
        # self._center = Vector(center) if center is not None else Vector()

    @classmethod
    def from_bounding_box(cls, bbox: BoundingBox) -> Rect:
        """Create a rect from a BoundingBox.

        Args:
            bbox: bounding box from which a rect is constructed

        Returns:
            Rect
        """
        return cls(bbox.width, bbox.height, 0., bbox.center)

    @classmethod
    def from_line(cls, line: Line, height: float):
        """Create a rect defined by a line a certain height

        Args:
            line (Line):
            height (float):

        Returns:
            Rect
        """
        direction = (line.end - line.start)
        width = direction.mag

        # is theta correct?
        return cls(width=width, height=height, center=line.center, theta=direction.angle_about_x_axis)

    def to_arc_spline(self) -> ArcSpline:
        vertices = np.append(self.corners, np.zeros(4)[:, np.newaxis], axis=1)

        return ArcSpline(vertices, True, self.description)

    @property
    def width(self) -> float:
        """Width of rect.

        Access:
            get

        Returns:
            float
        """
        start, end = self._rect.vertices[(0, 1), :2]
        return np.linalg.norm(start - end)

    @property
    def height(self) -> float:
        """Height of rect.

        Access:
            get

        Returns:
            float
        """
        start, end = self._rect.vertices[(1, 2), :2]
        return np.linalg.norm(start - end)

    @property
    def theta(self) -> float:
        """ Rotation angle of rect.

        Access:
            get

        Returns:
            float
        """
        return self._rot_angle

    @property
    def corners(self) -> List[Vector]:
        """Corner points of rect

        Access:
            get

        Returns:
            List[Vector]
        """
        return [Vector(vec) for vec in self._rect.vertices[:, :2]]

    def __repr__(self) -> str:
        return '{}(width={!r}, height={!r}, theta={!r}, center={!r})'.format(
            self.__class__.__name__, self.width, self.height, self.theta, self.center)

    @property
    def center(self) -> Vector:
        return self._rect.center

    @property
    def bounding_box(self) -> BoundingBox:
        return BoundingBox.from_points(self.corners)

    @property
    def is_closed(self) -> bool:
        return True

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._rect._impl_translate(trans_vec)

    def _impl_rotate(self, theta: float) -> None:
        self._rect._impl_rotate(theta)
        self._rot_angle += float(theta)

    def _impl_scale(self, fac: float) -> None:
        self._rect._impl_scale(fac)

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        # how das mirroring affect rot angle?
        # TODO: test this carefully
        ref_vec = self._rect.vertices[0, :2]

        self._rect._impl_mirror(mirror_axis)

        rot_angle = signed_angle_between(ref_vec, self._rect.vertices[0, :2])
        self._rot_angle += rot_angle




from typing import Optional

from fibomat.linalg import DimTransformable, DimVector, DimVectorLike, DimBoundingBox
from fibomat.shapes.shape import Shape
from fibomat.units import U_, has_length_dim, scale_factor


class DimShape(DimTransformable):
    def __init__(self, shape: Shape, unit: U_, description: Optional[str] = None):
        self._shape = shape

        if not isinstance(unit, U_):
            raise TypeError('unit must be of type pint.Unit (aka U_).')

        if not has_length_dim(unit):
            raise ValueError('unit must have dimension [lenght].')

        self._unit = unit

        super().__init__(description)

    def __repr__(self):
        return '{}(shape={!r}, unit={!r})'.format(self.__class__.__name__, self._shape, self._unit)

    @property
    def shape(self):
        return self._shape

    @property
    def unit(self):
        return self._unit

    @property
    def center(self) -> DimVector:
        return DimVector(self._shape.center * self._unit)

    @property
    def bounding_box(self) -> DimBoundingBox:
        return self._shape.bounding_box * self._unit

    def _impl_translate(self, trans_vec: DimVectorLike) -> None:
        trans_vec = DimVector(trans_vec)
        self._shape._impl_translate(scale_factor(self._unit, trans_vec.unit) * trans_vec.vector)

    def _impl_rotate(self, theta: float) -> None:
        self._shape._impl_rotate(theta)

    def _impl_scale(self, fac: float) -> None:
        self._shape._impl_scale(fac)

    def _impl_mirror(self, mirror_axis: DimVectorLike) -> None:
        mirror_axis = DimVector(mirror_axis)

        # vector length does not matter, mirror axis is normalized anyway.
        self._shape._impl_mirror(mirror_axis.vector)



from typing import Union, List
import pathlib
import math

import numpy as np

import ezdxf
import ezdxf.entities

from fibomat.shapes import Shape, Polyline, Polygon, Line, Spot, Circle, Ellipse, Rect
from fibomat.utils import PathLike


def _is_rect(vertices):
    if len(vertices) == 4:
        # check if the diagonals are equal
        if math.isclose(np.linalg.norm(vertices[0]-vertices[2]),
                        np.linalg.norm(vertices[1]-vertices[3])):
            return True
    return False


def _rect_from_vertices(vertices):

    # find the vertex with the minimum y value to get a base site of the rect.
    i_min = np.where(vertices[:, 1] == vertices[:, 1].min())[0]

    if len(i_min) == 1:
        # we found one, now check, if the left or right vertex will result in a
        # smaller rotation of the base site compared to x axis.
        base_1 = vertices[(i_min[0]+1) % 4] - vertices[i_min[0]]
        base_2 = vertices[i_min[0]] - vertices[(i_min[0]-1) % 4]

        if np.abs(np.arctan2(base_1[1], base_1[0])) < np.abs(np.arctan2(base_2[1], base_2[0])):
            base_start_vertex_i = i_min[0]
            base_end_vertex_i = (i_min[0]+1) % 4

            theta = np.arctan2(base_1[1], base_1[0])
        else:
            base_start_vertex_i = (i_min[0]-1) % 4
            base_end_vertex_i = i_min[0]

            theta = np.arctan2(base_2[1], base_2[0])

    elif len(i_min) == 2:
        # we got two, the base site is parallel to x axis
        base_start_vertex_i = i_min[0]
        base_end_vertex_i = i_min[1]

        if vertices[base_start_vertex_i, 0] > vertices[base_end_vertex_i, 0]:
            base_start_vertex_i, base_end_vertex_i = base_end_vertex_i, base_start_vertex_i

        base = vertices[base_end_vertex_i] - vertices[base_start_vertex_i]
        theta = np.arctan2(base[1], base[0])
    else:
        raise RuntimeError('Something went horribly wrong.')

    width = np.linalg.norm(vertices[base_start_vertex_i]-vertices[base_end_vertex_i])
    height = np.linalg.norm(vertices[base_end_vertex_i]-vertices[(base_end_vertex_i+1) % 4])

    center = np.mean(vertices, axis=0)

    return Rect(width=width, height=height, theta=theta, center=center)


# TODO: let vertices_raw be a generator object
def _polyline(ocs, vertices_raw, closed):
    vertices = np.empty(shape=(len(vertices_raw), 2))
    for i, vertex in enumerate(vertices_raw):
        vertices[i][0] = ocs.to_wcs(vertex)[0]
        vertices[i][1] = ocs.to_wcs(vertex)[1]

    if closed:
        if _is_rect(vertices):
            return _rect_from_vertices(vertices)
        return Polygon(vertices)

    # it could be possible that start and end are equal but it is not marked as
    # closed.
    if np.allclose(vertices[0], vertices[-1]):
        cropped_vertices = vertices[:-1]
        if _is_rect(cropped_vertices):
            return _rect_from_vertices(cropped_vertices)
        return Polygon(cropped_vertices)

    return Polyline(vertices)


def _ellipse(major, minor, center):
    major = np.array(major)
    minor = np.array(minor)

    angle_major = np.abs(np.arctan2(major[1], major[0]))

    if angle_major > np.pi/2:
        major *= -1.
        angle_major -= np.pi

    angle_minor = np.abs(np.arctan2(minor[1], minor[0]))
    if angle_minor > np.pi / 2:
        angle_minor *= -1.
        angle_major -= np.pi

    # if angle_major > np.pi/2:
    #     angle_major -= np.pi / 2
    #
    # if angle_minor > np.pi / 2:
    #     angle_minor -= np.pi / 2

    print(angle_major, angle_minor)

    if angle_major <= angle_minor:
        a = np.linalg.norm(major)
        b = np.linalg.norm(minor)

        theta = np.arctan2(major[1], major[0])

    else:
        a = np.linalg.norm(minor)
        b = np.linalg.norm(major)

        theta = np.arctan2(minor[1], minor[0])

    if theta > np.pi / 2:
        theta -= np.pi
    elif theta < -np.pi / 2:
        theta += np.pi

    return Ellipse(a, b, theta, center)


def shapes_from_dxf(file_path: PathLike) -> List[Shape]:
    """
    Reads a dxf file an convert the contained entities (shapes) to fibomat shapes.

    .. note:: z components of all entities are ignored.

    Args:
        file_path (pathlib.Path): path to file

    Returns:
        List[Shape]
    """
    doc = ezdxf.readfile(str(file_path))

    shapes = []

    # we try to get all objects from modelspace
    for obj in doc.modelspace():
        ocs = obj.ocs()
        dxftype = obj.dxftype()

        if dxftype == 'POINT':
            spot: ezdxf.entities.Point = obj
            shapes.append(Spot(spot.dxf.location[:2]))
        elif dxftype == 'LINE':
            line: ezdxf.entities.Line = obj
            shapes.append(Line(line.dxf.start[:2], line.dxf.end[:2]))
        elif dxftype == 'POLYLINE':
            poly: ezdxf.entities.Polyline = obj

            # maybe these to clauses are equivalent
            if not poly.is_2d_polyline or not poly.get_mode() == 'AcDb2dPolyline':
                raise RuntimeError('Polyline is not 2d.')

            if poly.has_arc:
                raise RuntimeError('Arcs are not supported in polyline.')

            vertices = list(poly.points())
            shapes.append(_polyline(ocs, vertices, poly.is_closed))
        elif dxftype == 'LWPOLYLINE':
            poly: ezdxf.entities.LWPolyline = obj

            if poly.has_arc:
                raise RuntimeError('Arcs are not supported in polyline.')

            vertices = list(poly.vertices())
            shapes.append(_polyline(ocs, vertices, poly.closed))
        elif dxftype == 'CIRCLE':
            circle: ezdxf.entities.Circle = obj
            shapes.append(
                Circle(
                    r=circle.dxf.radius,
                    center=ocs.to_wcs(circle.dxf.center)[:2])
            )
        elif dxftype == 'ELLIPSE':
            ellipse: ezdxf.entities.Ellipse = obj
            if not math.isclose(ellipse.dxf.start_param, 0.) or not math.isclose(ellipse.dxf.end_param, 2.* np.pi):
                raise RuntimeError("Elliptical arcs are not supported.")

            shapes.append(_ellipse(ellipse.dxf.major_axis[:2], ellipse.minor_axis[:2], ellipse.dxf.center[:2]))
        else:
            raise RuntimeError(
                f'Encountered (currently) unsupported shape {obj.dxftype()}'
            )

    return shapes



from typing import Tuple, List, Optional

import numpy as np

from fibomat.rasterizedpattern import RasterizedPattern
from fibomat.shapes.rasterizedpoints import RasterizedPoints
from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.units import LengthUnit, TimeUnit, scale_to, has_length_dim, QuantityType, scale_factor, LengthQuantity
from fibomat.shapes.arc_spline import ArcSplineCompatible, ArcSpline
from fibomat.shapes import DimShape
from fibomat.mill import Mill
from fibomat.curve_tools import inflate, deflate, rasterize, smooth
from fibomat.curve_tools.smooth import NonSmoothableError
from fibomat.optimize import optimize_curve
from fibomat.raster_styles.scansequence import ScanSequence, _apply_scan_sequence
from fibomat.raster_styles.one_d.curve import Curve
from fibomat.raster_styles.zero_d.prerasterized import PreRasterized


def _print_optimize_result(alpha, flux_matrix, nominal_flux):
    print('mean_alpha =', np.mean(alpha))
    print('std_alpha =', np.std(alpha))
    print('min_alpha = ', np.min(alpha))
    print('max_alpha = ', np.max(alpha))

    flux = flux_matrix @ alpha
    print('mean_flux =', np.mean(flux))
    print('std_flux =', np.std(flux))
    print('min_flux = ', np.min(flux))
    print('max_flux = ', np.max(flux))
    print('nominal_flux =', nominal_flux)


class ContourParallel(RasterStyle):

    _inwards = 'inwards'
    _outwards = 'outwards'

    def __init__(
        self,
        offset_pitch: LengthQuantity,
        offset_direction: str,
        start_direction: str,
        scan_sequence: ScanSequence,
        line_style: RasterStyle,
        include_original_curve: bool,
        optimize: bool,
        *,
        offset_steps: Optional[int] = None,
        offset_distance: Optional[LengthQuantity] = None,
        smooth_radius: Optional[LengthQuantity] = None
    ):
        self._scan_sequence = scan_sequence

        self._line_style = line_style

        if not has_length_dim(offset_pitch):
            raise ValueError('offset_pitch must have [length] as dimension')

        if offset_pitch <= 0.:
            raise ValueError('offset_pitch <= 0.')
        self._offset_pitch = offset_pitch

        if offset_direction not in [self._inwards, self._outwards]:
            raise ValueError(f'offset_direction must be "{self._inwards}" or "{self._outwards}".')

        self._offset_direction = offset_direction
        self._offset_steps = offset_steps
        self._offset_distance = offset_distance

        if start_direction not in [self._inwards, self._outwards]:
            raise ValueError(f'start_direction must be "{self._inwards}" or "{self._outwards}".')
        self._start_direction = start_direction

        self._include_original_curve = include_original_curve

        self._smooth_radius = smooth_radius
        self._optimize = optimize

    @property
    def dimension(self) -> int:
        return 2

    def _do_smooth(self, spline: ArcSpline, curve_unit: LengthUnit):
        if self._smooth_radius:
            try:
                return smooth(spline, self._smooth_radius.to(curve_unit).m)
            except NonSmoothableError as e:
                print('Warning: could not smooth spline.')
                return spline
        else:
            return spline

    def _gen_offset_curves(self, base_curve: ArcSpline, curve_unit: LengthUnit) -> List[ArcSpline]:

        curves = []
        if self._include_original_curve:
            curves.append(base_curve)

        if self._offset_direction == self._inwards:
            curves.extend(deflate(
                base_curve,
                self._offset_pitch.to(curve_unit).m,
                n_steps=self._offset_steps,
                distance=self._offset_distance.to(curve_unit).m if self._offset_distance else None
            ))

            if self._start_direction == self._outwards:
                curves = list(reversed(curves))

        else:
            curves.extend(inflate(
                base_curve,
                self._offset_pitch.to(curve_unit).m,
                n_steps=self._offset_steps,
                distance=self._offset_distance.to(curve_unit).m if self._offset_distance else None
            ))

            if self._start_direction == self._inwards:
                curves = list(reversed(curves))

        return [self._do_smooth(curve, curve_unit) for curve in curves]

    def rasterize(
        self,
        dim_shape: DimShape,
        mill: Mill,
        out_length_unit: LengthUnit,
        out_time_unit: TimeUnit
    ) -> RasterizedPattern:

        # dim_shape = DimObj.create(dim_shape)

        offsetted_curves = self._gen_offset_curves(dim_shape.obj.to_arc_spline(), dim_shape.unit)

        if not self._optimize:
            return _apply_scan_sequence(
                filling_rows=offsetted_curves,
                filling_rows_length_unit=dim_shape.unit,
                scan_sequence=self._scan_sequence,
                line_style=self._line_style,
                mill=mill,
                out_length_unit=out_length_unit,
                out_time_unit=out_time_unit
            )
        else:
            if not (
                isinstance(self._line_style, Curve) and self._line_style.scan_sequence == ScanSequence.CONSECUTIVE
            ):
                raise RuntimeError(
                    'Only one_d.Curve with ScanSequence.CONSECUTIVE scan sequence is supported if optimize == True.'
                )

            nominal_flux = mill.beam.nominal_flux_per_spot_on_line(self._line_style.pitch).to('ions / nm**2 / µs')
            hint = None

            rasterized_curves = []

            for i_curve, curve in enumerate(offsetted_curves):
                print(i_curve, '/', len(offsetted_curves) - 1)

                dwell_points, hint, flux_matrix = optimize_curve(
                    (curve, dim_shape.unit), self._line_style.pitch, mill.beam, nominal_flux, hint, info=True
                )
                _print_optimize_result(dwell_points[:, 2], flux_matrix, nominal_flux)

                rasterized_curves.append(RasterizedPoints(dwell_points, dim_shape.obj.is_closed))

            return _apply_scan_sequence(
                filling_rows=rasterized_curves,
                filling_rows_length_unit=dim_shape.unit,
                scan_sequence=self._scan_sequence,
                line_style=PreRasterized(),
                mill=mill,
                out_length_unit=out_length_unit,
                out_time_unit=out_time_unit
            )



from typing import Tuple, Optional, Callable, Union

import numpy as np
# import open3d as o3d

import scipy.spatial as spatial
import scipy.interpolate as interpolate

from fibomat.shapes import ArcSpline
from fibomat.mill.ionbeam import GaussBeam, IonBeam
from fibomat.units import Q_, UnitType, U_, scale_factor, QuantityType, LengthQuantity, scale_to, LengthUnit
# from fibomat.sample import Sample
from fibomat.curve_tools import rasterize, smooth, deflate
from fibomat.dimensioned_object import DimObjLike, DimObj


_FLUX_UNIT = U_('ions / nm**2 / µs')


def _create_adjacency_matrix(dwell_points: np.ndarray, length_unit: LengthUnit, beam: IonBeam):
    beam_cutoff = (beam.std * beam._beam_cutoff).to(length_unit).m

    spatial_tree = spatial.cKDTree(dwell_points[:, :2])

    # adjacent_point_indices = []
    #
    # for i in range(len(dwell_points)):
    #     adjacent_point_indices.append(spatial_tree.query_ball_point(dwell_points[i, :2], beam_cutoff))

    adjacent_point_indices = np.zeros(shape=(len(dwell_points), len(dwell_points)), dtype=float)

    for i in range(len(dwell_points)):
        indices = spatial_tree.query_ball_point(dwell_points[i, :2], beam_cutoff)
        adjacent_point_indices[i, indices] = 1.


    # f = plt.matshow(adjacent_point_indices, cmap='Greys')
    # plt.colorbar(f)
    # plt.show()

    return adjacent_point_indices


def _create_flux_matrix(
    dwell_points: np.ndarray,
    connectivity_matrix: np.ndarray,
    beam: IonBeam,
):
    global _FLUX_UNIT

    flux_matrix = np.empty((len(dwell_points), len(dwell_points)), dtype=float)

    for i in range(len(dwell_points)):
        connectivity = connectivity_matrix[i]
        flux, flux_unit = beam.flux_at(
            dwell_points[i, :2],
            dwell_points[:, :2],
            U_('µm')
        )

        scale = scale_factor(_FLUX_UNIT, flux_unit)

        flux_matrix[i] = connectivity * dwell_points[:, 2] * scale * flux
        # connectivity = connectivity_matrix[i]
        # dose[i] = np.sum(
        #         dwell_points[connectivity, 2] * beam(
        #             dwell_points[i, 0], dwell_points[i, 1],
        #             dwell_points[connectivity[i], 0], dwell_points[connectivity[i], 1]
        #         )
        # )
    return flux_matrix


# https://stackoverflow.com/a/43259123
# def split_above_threshold(signal, threshold, above_or_below):
#     if above_or_below:
#         mask = np.concatenate(([False], signal > threshold, [False]))
#     else:
#         mask = np.concatenate(([False], signal < threshold, [False]))
#     idx = np.flatnonzero(mask[1:] != mask[:-1])
#     return [(idx[i], idx[i+1]) for i in range(0, len(idx), 2)]


def optimize_rasterized(dwell_points, length_unit: LengthUnit, beam: IonBeam, nominal_flux: QuantityType, hint: Optional[Callable] = None):
    _MAX_ITER = 1000

    adj_matrix = _create_adjacency_matrix(dwell_points, length_unit, beam)
    inv_mean_connectivity = (1/np.sum(adj_matrix, axis=1))
    flux_matrix = _create_flux_matrix(dwell_points, adj_matrix, beam)

    initial_flux = flux_matrix @ np.ones(len(dwell_points))

    if hint:
        alpha = np.asarray(hint(np.linspace(0, 1, len(dwell_points))))
    else:
        alpha = np.full_like(initial_flux, 1.)

    max_initial_flux = np.max(initial_flux)

    alpha_max = 1.2

    # if not hint:
    # for threshold_factor in np.arange(.2, 1, 0.01):
    #     for _ in range(_MAX_ITER):
    #         # intervals = split_above_threshold(flux_matrix @ alpha, threshold_factor*max_initial_flux, False)  # .99 * nominal_flux.m
    #         # print(len(intervals))
    #
    #         idx = np.argwhere((flux_matrix @ alpha) < threshold_factor*max_initial_flux)
    #
    #         if len(idx) == 0:
    #             break
    #
    #         if np.allclose(alpha[idx], alpha_max):
    #             break
    #
    #         alpha[idx] += inv_mean_connectivity[idx]
    #
    #         alpha[alpha > alpha_max] = alpha_max

    nominal_flux = scale_to(_FLUX_UNIT, nominal_flux)

    for threshold_factor in reversed(np.arange(1, 2, 0.05)): # [1.6, 1.5, 1.4, 1.3, 1.2, 1.1, 1.05, 1.01, 1.]:
        # print(threshold_factor * nominal_flux.m)
        for _ in range(_MAX_ITER):

            idx = np.argwhere((flux_matrix @ alpha) > threshold_factor * nominal_flux)

            if len(idx) == 0:
                break

            if np.allclose(alpha[idx], 0.):
                break

            alpha[idx] -= inv_mean_connectivity[idx] #

            alpha[alpha < 0] = 0.

    dwell_points[:, 2] = alpha
    if len(dwell_points) > 1:
        hint = interpolate.interp1d(np.linspace(0, 1, len(dwell_points)), alpha)
    else:
        hint = None

    return dwell_points, hint, flux_matrix

import numba


@numba.jit(nopython=True, fastmath=True)
def _impl(flux_0, alpha, conn, threshold):
    last_std = np.std(flux_0 @ alpha)

    ref_step = 0
    step = 0.1

    while last_std > threshold:
        new_alpha = alpha.copy()
        flux_t = flux_0 @ alpha
        for i in range(len(alpha)):
            new_alpha[i] = alpha[i]  # - step * (14 - flux_t[i])
            new_alpha[i] -= step * np.sum(flux_t[i] - flux_t[conn[i]])

        new_alpha[new_alpha < 0.] = 0.
        new_alpha[new_alpha > 3.] = 3.
        std = np.std(flux_0 @ new_alpha)
        # print(std)

        if last_std < std:
            # raise RuntimeError('WARNING: last_std < std:')
            print('WARNING: last_std < std:')

            ref_step += 1
            step /= 10

            if ref_step > 4:
                print('WARNING: last_std < std. BREAKING NOW.')
                break

            continue

        last_std = std
        alpha = new_alpha

    return alpha, last_std


# https://www.labri.fr/perso/ejeannot/publications/europar06.pdf
# http://www.dartmouth.edu/~gvc/Cybenko_JPDP.pdf
def optimze_rasterized_second_try(dwell_points, length_unit: LengthUnit, beam: IonBeam, nominal_flux: QuantityType, hint: Optional[Callable] = None):

    _MAX_ITER = 1000

    threshold = 0.025

    adj_matrix = _create_adjacency_matrix(dwell_points, length_unit, beam)
    flux_matrix = _create_flux_matrix(dwell_points, adj_matrix, beam)

    # # algorithm will converge to mean value. hence, we shift the mean of initial flux to the nominal flux
    # # flux_matrix += nominal_flux.to(_FLUX_UNIT).m - np.mean(flux_matrix)
    # print(np.sum(flux_matrix, axis=1) - nominal_flux.to(_FLUX_UNIT).m)
    # print(np.sum(flux_matrix + (nominal_flux.to(_FLUX_UNIT).m - np.mean(flux_matrix)) / np.sum(adj_matrix) * adj_matrix, axis=1))
    #
    # print(np.mean(np.sum(flux_matrix, axis=1)))
    #

    flux_0 = flux_matrix
    if hint:
        alpha = hint(np.linspace(0, 1, len(flux_matrix)))
    else:
        alpha = np.ones(len(flux_matrix))

    # mean_flux = np.mean(flux_0 @ alpha)
    # alpha *= nominal_flux.to(_FLUX_UNIT).m / mean_flux

    conn = flux_matrix > 0
    # conn = adj_matrix

    # last_std = np.std(flux_0 @ alpha)
    #
    # ref_step = 0
    # step = 0.01
    #
    # while last_std > threshold:
    #     new_alpha = alpha.copy()
    #     flux_t = flux_0 @ alpha
    #     for i in range(len(alpha)):
    #         new_alpha[i] = alpha[i]
    #         new_alpha[i] -= step * np.sum(flux_t[i] - flux_t[conn[i]])
    #
    #     new_alpha[new_alpha < 0.] = 0.
    #     new_alpha[new_alpha > 1.5] = 1.5
    #     std = np.std(flux_0 @ new_alpha)
    #     print(std)
    #
    #     if last_std < std:
    #         # raise RuntimeError('WARNING: last_std < std:')
    #         print('WARNING: last_std < std:')
    #
    #         ref_step += 1
    #         step /= 10
    #
    #         if ref_step > 3:
    #             print('WARNING: last_std < std. BREAKING NOW.')
    #             break
    #
    #         continue
    #
    #     last_std = std
    #     alpha = new_alpha

    alpha, last_std = _impl(flux_0, alpha, conn, threshold)

    if last_std > threshold:
        # raise RuntimeError('last_std > threshold. Optimization did not converge.')
        print('WARNING: last_std > threshold. Optimization did not converge.')

    # TODO: NOT GOOD!
    alpha[alpha > 3.] = 1.5

    mean_flux = np.mean(flux_matrix @ alpha)
    print('mean_flux:', mean_flux)

    print('fac', mean_flux / nominal_flux.to(_FLUX_UNIT).m )

    # scale alpha so that flux mean is nominal_flux
    dwell_points[:, 2] = alpha * nominal_flux.to(_FLUX_UNIT).m / mean_flux
    if len(dwell_points) > 1:
        hint = interpolate.interp1d(np.linspace(0, 1, len(dwell_points)), alpha)
    else:
        hint = None

    return dwell_points, hint, flux_matrix


def optimze_rasterized_hybrid(dwell_points, length_unit: LengthUnit, beam: IonBeam, nominal_flux: QuantityType, hint: Optional[Callable] = None):
    _MAX_ITER = 1000

    adj_matrix = _create_adjacency_matrix(dwell_points, length_unit, beam)
    inv_mean_connectivity = (1/np.sum(adj_matrix, axis=1))

    flux_matrix = _create_flux_matrix(dwell_points, adj_matrix, beam)

    initial_flux = flux_matrix @ np.ones(len(dwell_points))

    if hint:
        alpha = np.asarray(hint(np.linspace(0, 1, len(dwell_points))))
    else:
        # alpha = np.full_like(initial_flux, 1.)
        alpha = np.array(dwell_points[:, 2])

    nominal_flux = scale_to(_FLUX_UNIT, nominal_flux)

    # first pass
    for threshold_factor in reversed(np.arange(1, 2, 0.05)): # [1.6, 1.5, 1.4, 1.3, 1.2, 1.1, 1.05, 1.01, 1.]:
        # print(threshold_factor)
        for _ in range(_MAX_ITER):

            idx = np.argwhere((flux_matrix @ alpha) > threshold_factor * nominal_flux)

            if len(idx) == 0:
                break

            if np.allclose(alpha[idx], 0.):
                break

            alpha[idx] -= inv_mean_connectivity[idx] #

            alpha[alpha < 0] = 0.

    # second pass
    threshold = 0.025

    alpha, last_std = _impl(flux_matrix, alpha, flux_matrix > 0, threshold)

    if last_std > threshold:
        # raise RuntimeError('last_std > threshold. Optimization did not converge.')
        print('WARNING: last_std > threshold. Optimization did not converge.')

    # TODO: NOT GOOD! WARN!
    alpha[alpha > 1.5] = 1.5

    # scale alpha so that flux mean is nominal_flux
    dwell_points[:, 2] = alpha * nominal_flux / np.mean(flux_matrix @ alpha)
    if len(dwell_points) > 1:
        hint = interpolate.interp1d(np.linspace(0, 1, len(dwell_points)), alpha)
    else:
        hint = None

    return dwell_points, hint, flux_matrix




def optimize_curve(
    dim_spline: DimObjLike[ArcSpline, LengthUnit],
    pitch: LengthQuantity,
    beam: IonBeam,
    nominal_flux: QuantityType,
    hint: Optional[Callable] = None,
    info: bool = False
) -> Union[np.ndarray, Tuple[np.ndarray, Callable, np.ndarray]]:
    """Optimize the the dwell points an a curve so every point has an ion flux of `nominal_flux`.

    Args:
        dim_spline (DimObjLike[ArcSpline, LengthUnit]): curve with unit to be rasterized and optimized
        pitch (LengthQuantity): rasterization pitch
        beam (IonBeam): beam
        nominal_flux (QuantityType): nominal ion flux per spot
        hint (Callable, optional):
            a function f: [0, 1] -> R mapping rasterized points indizes (which are itself mappedto [0, 1] to dwell time
            multiplicands.
        info (bool): if True, dwell_points, hint and flux_matrix are returned.

    Returns:

    """

    dim_spline = DimObj.create(dim_spline)

    dwell_points = np.array(rasterize(dim_spline.obj, scale_to(dim_spline.unit, pitch)).dwell_points)

    dwell_points, hint, flux_matrix = optimze_rasterized_hybrid(dwell_points, dim_spline.unit, beam, nominal_flux, hint)
    # dwell_points, hint, flux_matrix = optimze_rasterized_second_try(dwell_points, dim_spline.unit, beam, nominal_flux, hint)

    if info:
        return dwell_points, hint, flux_matrix

    return dwell_points

# from typing import List, Tuple
#
# import numpy as np
# import cvxpy as cp
# from scipy import spatial
#
# from fibomat import units, rasterizing
#
# from fibomat.mill.gaussbeam import GaussBeam


# def _create_adjacency_matrix(dwell_points: np.ndarray, beam: GaussBeam):
#     beam_cutoff = (beam.sigma * beam.BEAM_CUTOFF).to('nm').m
#
#     spatial_tree = spatial.cKDTree(dwell_points[:, :2])
#
#     # adjacent_point_indices = []
#     #
#     # for i in range(len(dwell_points)):
#     #     adjacent_point_indices.append(spatial_tree.query_ball_point(dwell_points[i, :2], beam_cutoff))
#
#     adjacent_point_indices = np.zeros(shape=(len(dwell_points), len(dwell_points)), dtype=float)
#
#     for i in range(len(dwell_points)):
#         indices = spatial_tree.query_ball_point(dwell_points[i, :2], beam_cutoff)
#         adjacent_point_indices[i, indices] = 1.
#
#     import matplotlib.pyplot as plt
#     import matplotlib
#     matplotlib.use('Qt5Cairo')
#
#     print(adjacent_point_indices)
#     f = plt.matshow(adjacent_point_indices * 10000, cmap='Greys')
#     plt.colorbar(f)
#     plt.show()
#
#     return adjacent_point_indices
#
#
# def _create_dose_matrix(
#         dwell_points: np.ndarray,
#         connectivity_matrix: np.ndarray,
#         beam: GaussBeam,
#         flux_unit: units.UnitType
# ):
#     dose = np.empty((len(dwell_points), len(dwell_points)), dtype=float)
#
#     for i in range(len(dwell_points)):
#         connectivity = connectivity_matrix[i]
#         dose[i] = connectivity * dwell_points[:, 2] * beam.flux_at(
#             dwell_points[i, 0], dwell_points[i, 1],
#             dwell_points[:, 0], dwell_points[:, 1],
#             flux_unit
#         )
#         # connectivity = connectivity_matrix[i]
#         # dose[i] = np.sum(
#         #         dwell_points[connectivity, 2] * beam(
#         #             dwell_points[i, 0], dwell_points[i, 1],
#         #             dwell_points[connectivity[i], 0], dwell_points[connectivity[i], 1]
#         #         )
#         # )
#     return dose
#
#
# def optimize_curve_dwell_times(
#     dim_curves: List[Tuple[rasterizing.RasterizedPoints, units.LengthUnit]],
#     dwell_times_per_curve: List[units.TimeQuantity],
#     # curve_units: Tuple[units.UnitType, units.UnitType],
#     nominal_fluxes_per_curve: List[units.QuantityType],
#     beam: GaussBeam,
#     mode: str = 'min_dose_var',
#     plot_solution: bool = False
# ) -> List[rasterizing.RasterizedPoints]:
#     """
#     This function optimizes the dwell times of the given rasterized curves.
#
#     ...
#
#     .. note:: curves are modified!
#
#     Args:
#         dim_curves:
#         dwell_times_per_curve:
#         nominal_fluxes_per_curve:
#         beam:
#         mode:
#         plot_solution:
#
#     Returns:
#
#     """
#     if len(dim_curves) < 1:
#         raise RuntimeError
#     if len(dim_curves) != len(nominal_fluxes_per_curve):
#         raise RuntimeError
#     if len(dim_curves) != len(dwell_times_per_curve):
#         raise RuntimeError
#
#     flux_unit = units.U_('ions / µm**2 / µs')
#
#     # prepare curves
#     n_dwell_points = sum([len(curve.dwell_points) for curve, _ in dim_curves])
#     dwell_points = np.empty(shape=(n_dwell_points, 3), dtype=float)
#     nominal_dose_vector = np.empty(shape=(n_dwell_points,), dtype=float)
#
#     offset = 0
#     for (curve, curve_unit), dwell_time, nominal_flux in zip(dim_curves, dwell_times_per_curve, nominal_fluxes_per_curve):
#         # if curve_unit != units.U_('µm'):
#
#         point_range = slice(offset, offset+len(curve.dwell_points))
#
#         dwell_points[point_range, :2] = units.scale_factor(units.U_('µm'), curve_unit) * curve.dwell_points[:, :2]
#         dwell_points[point_range, 2] = units.scale_to(units.U_('µs'), dwell_time)
#
#         nominal_dose_vector[point_range] = nominal_flux.to(flux_unit).m
#
#         nominal_dose_vector *= dwell_points[:, 2]
#
#         offset += len(curve.dwell_points)
#
#     # precalculate adjacency matrix ("which other spots influence spot i")  and dose matrix ("which dose contributes
#     # spot j to spot j")
#     # dose_unit = units.U_('ions / µm**2 / µs')
#     adjacency_matrix = _create_adjacency_matrix(dwell_points, beam)
#     dose_matrix = _create_dose_matrix(dwell_points, adjacency_matrix, beam, flux_unit)
#
#     # return dose_matrix @ np.ones(len(dwell_points), dtype=float)
#
#     # # normalize doses
#     # dose_scaling = 1 / np.mean(nominal_dose_vector)
#     # dose_matrix *= dose_scaling
#     # nominal_dose_vector *= dose_scaling
#
#     # run optimization
#     n_points = len(dwell_points)
#     alpha = cp.Variable(n_points)
#     alpha.value = np.ones(len(dwell_points), dtype=float)
#
#     if mode == 'min_dose_var':
#         cost = cp.sum_squares(dose_matrix @ alpha - nominal_dose_vector) # + cp.sum_squares(alpha - 1.)
#         constrains = [alpha >= 0.]
#     elif mode == 'min_alpha_var':
#         # cost = cp.sum_squares((alpha - cp.sum(alpha) / n_points) / n_points)
#         # cost = cp.norm2((alpha - cp.sum(alpha) / n_points)) / n_points
#         ones = np.ones(len(dwell_points), dtype=float)
#         cost = cp.norm2(alpha - 1.)
#         constrains = [cp.abs(dose_matrix @ alpha - nominal_dose_vector) <= 0.01 * nominal_dose_vector, alpha >= 0.]
#     else:
#         raise RuntimeError(f'Unknown mode "{mode}"')
#
#     prob = cp.Problem(cp.Minimize(cost), constrains)
#
#     print(prob.is_dcp())
#
#     res = prob.solve(warm_start=True)
#     if prob.status != cp.OPTIMAL:
#         print(res)
#         print(prob.solver_stats.solver_name)
#         raise RuntimeError("Solver did not converge!")
#
#     # save optimized dwell times and restore original scaling
#     alpha_sol = alpha.value
#
#     # import scipy.optimize
#     #
#     # def f(alpha):
#     #     return dose_matrix @ alpha - nominal_dose_vector
#     #
#     #
#     # def f_jac(alpha):
#     #     return dose_matrix
#     #
#     # alpha_0 = np.ones(len(dwell_points), dtype=float)
#     #
#     # sol = scipy.optimize.least_squares(f, alpha_0, bounds=(0, np.inf), jac=f_jac)
#     #
#     # print(sol)
#     #
#     # alpha_sol = sol.x
#
#
#     i_dwell_point = 0
#     for i, (curve, _) in enumerate(dim_curves):
#         i_range = slice(i_dwell_point, i_dwell_point+len(curve.dwell_points))
#         # print(curve.dwell_points[:, 2], alpha_sol[i_range], '\n')
#         curve.dwell_points[:, 2] = alpha_sol[i_range]
#         i_dwell_point += len(curve.dwell_points)
#
#     if plot_solution:
#         import matplotlib.pyplot as plt
#         import matplotlib
#         matplotlib.use('Qt5Cairo')
#
#         # https://nbviewer.jupyter.org/github/dpsanders/matplotlib-examples/blob/master/colorline.ipynb
#         ones = np.ones(n_points, dtype=float)
#
#         plt.matshow(adjacency_matrix * 10000, cmap='Greys')
#         print(np.where(adjacency_matrix > 0.))
#         plt.savefig('foo.png',dpi=1000)
#         plt.show()
#
#         doses_before_opt = dose_matrix @ ones
#         dose_diffs_before = dose_matrix @ ones - nominal_dose_vector
#         doses_after_opt = dose_matrix @ alpha.value
#
#         fig, axs = plt.subplots(3, 2, constrained_layout=True)
#
#         data_sources = [nominal_dose_vector, doses_before_opt, doses_after_opt] # dose_diffs_before,
#         vmin = min([np.min(source) for source in data_sources])
#         vmax = max([np.max(source) for source in data_sources])
#
#         # plot_nominal_doses = axs[0, 0].scatter(
#         #     dwell_points[:, 0], dwell_points[:, 1], c=nominal_dose_vector, vmin=vmin, vmax=vmax
#         # )
#         # fig.colorbar(plot_nominal_doses, ax=axs[0, 0])
#         # axs[0, 0].set_aspect('equal')
#
#         axs[0, 0].imshow(dose_matrix)
#
#         plot_doses_bef_opt = axs[0, 1].scatter(dwell_points[:, 0], dwell_points[:, 1], c=doses_before_opt, vmin=vmin, vmax=vmax)
#         axs[0, 1].set_xlabel('x / µm')
#         axs[0, 1].set_ylabel('y / µm')
#         cbar = fig.colorbar(plot_doses_bef_opt, ax=axs[0, 1])
#         cbar.ax.set_ylabel('dose multiplier')
#         axs[0, 1].set_aspect('equal')
#
#         plot_doses_after_opt = axs[1, 1].scatter(dwell_points[:, 0], dwell_points[:, 1], c=doses_after_opt, vmin=vmin, vmax=vmax)
#         cbar = fig.colorbar(plot_doses_after_opt, ax=axs[1, 1])
#         axs[1, 1].set_xlabel('x / µm')
#         axs[1, 1].set_ylabel('y / µm')
#         cbar.ax.set_ylabel('dose multiplier')
#         axs[1, 1].set_aspect('equal')
#
#         # adjacency_matrix.dtype = float
#         # adjacency_matrix[adjacency_matrix == 0] = None
#
#         axs[1, 0].hist(doses_after_opt, bins=100)
#
#         axs[2, 0].hist(dwell_points[:, 2] * alpha_sol, bins=100)
#
#         # data_sources = [nominal_dose_vector, doses_before_opt, dose_diffs_before, doses_after_opt]
#         # vmin = min([np.min(source) for source in data_sources])
#         # vmax = max([np.max(source) for source in data_sources])
#         #
#         #
#         # axs[1, 0].scatter(dwell_points[:, 0], dwell_points[:, 1], c=doses_after_opt, vmin=vmin, vmax=vmax)
#         # axs[1, 1].plot(np.arange(len(dwell_points)), nominal_dose_vector - doses_after_opt)
#         # fig.colorbar(plot_nominal_doses, ax=axs.flat)
#
#         plt.show()
#
#     return dim_curves



"""Provides the :class:`Line` class."""
from __future__ import annotations
from typing import Optional, List

from fibomat.linalg import Vector, VectorLike, BoundingBox
from fibomat.shapes import shape
from fibomat.shapes.line import Line


class LineNonContinuous(shape.Shape):
    def __init__(self, line_segments: List[Line]):
        super().__init__()
        self._segments = line_segments

    def __repr__(self):
        raise NotImplementedError

    @property
    def segments(self):
        return self._segments

    @property
    def is_closed(self) -> bool:
        return False

    @property
    def bounding_box(self) -> BoundingBox:
        raise NotImplementedError

    @property
    def center(self) -> Vector:
        raise NotImplementedError

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        raise NotImplementedError

    def _impl_rotate(self, theta: float) -> None:
        raise NotImplementedError

    def _impl_scale(self, fac: float) -> None:
        raise NotImplementedError

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        raise NotImplementedError



from typing import Tuple, List
import enum

import numpy as np

from fibomat.shapes import Shape
from fibomat.units import LengthUnit, TimeUnit
from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.mill import Mill
from fibomat.rasterizedpattern import RasterizedPattern


@enum.unique
class ScanSequence(enum.Enum):
    CONSECUTIVE = 'consecutive'
    BACKSTITCH = 'backstitch'
    BACK_AND_FORTH = 'back_and_forth'
    SERPENTINE = 'serpentine'
    DOUBLE_SERPENTINE = 'double_serpentine'
    DOUBLE_SERPENTINE_SAME_PATH = 'double_serpentine_same_path'
    CROSSECTION = 'crossection'


def _make_scan_index_sequence(
    n_lines: int, repeats: int, scan_sequence: ScanSequence
) -> Tuple[np.ndarray, Tuple[bool, np.ndarray]]:
    consecutive = np.arange(n_lines)

    if scan_sequence != ScanSequence.CROSSECTION:
        index_sequence = np.empty(n_lines * repeats, dtype=consecutive.dtype)
    else:
        index_sequence = consecutive
    # line_directions = np.empty(n_lines * repeats, dtype=bool)

    if scan_sequence == ScanSequence.CONSECUTIVE:
        for i in range(repeats):
            index_sequence[i*n_lines:(i + 1) * n_lines] = consecutive
        line_directions = True
    elif scan_sequence == ScanSequence.BACKSTITCH:
        for i in range(0, n_lines - 1, 2):
            index_sequence[i], index_sequence[i + 1] = (
                consecutive[i + 1], consecutive[i]
            )
        if n_lines % 2 == 1:
            index_sequence[n_lines-1] = n_lines-1
        for i in range(1, repeats):
            index_sequence[i*n_lines:(i + 1) * n_lines] = index_sequence[0:n_lines]
        line_directions = True
    elif scan_sequence == ScanSequence.CROSSECTION:
        # for i in range(n_lines):
        #     index_sequence[i*repeats:(i + 1) * repeats] = i
        line_directions = True
    elif scan_sequence == ScanSequence.SERPENTINE:
        line_directions = np.empty(n_lines * repeats, dtype=bool)
        for i in range(repeats):
            index_sequence[i * n_lines:(i + 1) * n_lines] = consecutive
            line_directions[i * n_lines:(i + 1) * n_lines:2] = True
            line_directions[i * n_lines + 1:(i + 1) * n_lines:2] = False
    elif scan_sequence == ScanSequence.DOUBLE_SERPENTINE:
        line_directions = np.empty(n_lines * repeats, dtype=bool)
        direction = True
        for i in range(repeats):
            if direction:
                index_sequence[i*n_lines:(i + 1) * n_lines] = consecutive
            else:
                index_sequence[i*n_lines:(i + 1) * n_lines] = consecutive[::-1]
            line_directions[i * n_lines:(i + 1) * n_lines:2] = True
            line_directions[i * n_lines + 1:(i + 1) * n_lines:2] = False
            direction = not direction
    elif scan_sequence == ScanSequence.DOUBLE_SERPENTINE_SAME_PATH:
        direction = True
        for i in range(repeats):
            if direction:
                index_sequence[i*n_lines:(i + 1) * n_lines] = consecutive
            else:
                index_sequence[i*n_lines:(i + 1) * n_lines] = consecutive[::-1]
            direction = not direction

        line_directions = True
    else:
        raise ValueError('Scan sequence not supported.')

    if isinstance(line_directions, bool):
        line_directions = np.full(n_lines * repeats, line_directions, dtype=bool)

    return index_sequence, line_directions


def _apply_scan_sequence(
    filling_rows: List[Shape],
    filling_rows_length_unit: LengthUnit,
    scan_sequence: ScanSequence,
    line_style: RasterStyle,
    mill: Mill,
    out_length_unit: LengthUnit,
    out_time_unit: TimeUnit
):
    if scan_sequence == ScanSequence.CROSSECTION:
        line_mill = mill
    else:
        line_mill = Mill(mill.dwell_time, repeats=1)

    rasterized_lines = []

    for row in filling_rows:
        rasterized_lines.append(
            line_style.rasterize(
                row * filling_rows_length_unit,
                mill=line_mill,
                out_length_unit=out_length_unit,
                out_time_unit=out_time_unit
            ).dwell_points
        )

    scan_lines = []

    for index, direction in zip(*_make_scan_index_sequence(len(rasterized_lines), mill.repeats, scan_sequence)):
        if not direction:
            scan_lines.append(rasterized_lines[index][::-1])
        else:
            scan_lines.append(rasterized_lines[index])

    return RasterizedPattern(np.concatenate(scan_lines), out_length_unit, out_time_unit)



"""
Provides the :class:`BackendBase` class.
"""
from typing import Dict, Callable, Type
import inspect

from fibomat.site import Site
from fibomat.pattern import Pattern
from fibomat import shapes
from fibomat.utils import PathLike
from fibomat.backend.backendbasemeta import BackendBaseMeta
from fibomat.rasterizedpattern import RasterizedPattern

from fibomat import layout


class ShapeNotSupportedError(TypeError):
    """Exception is used in :class:`BackendBase` for (all) unimplemented shape
    methods.
    """


def shape_type(type_: Type[shapes.Shape]) -> Callable:
    """Function is used as a decorator in :class:`BackendBase` to mark shape
    methods.
    See :class:`BackendBase` class and its implementation for details.

    .. note:: This decorator is only needed if a custom shape type is
              introduced and should be made available to the backend

    Args:
        type_ (Type[Shape]): Type of shape.

    Returns:
        Callable: Decorator
    """
    def decorator(func):
        # pylint: disable=protected-access
        func._shape_type = type_
        func._abstract_shape_method = True
        return func

    return decorator


class BackendBase(metaclass=BackendBaseMeta):
    """Base class for any backend. For all default shapes (see ...), a method
    stub is implemented.

    TODO: UPDATE THIS!

    Use :class:`BackendBase` as base class for a custom backend and implement
    all shape methods with should be supported and ignore the rest. If any
    other new shape is introduced and should be supported by the custom
    backend, use the following example as a base structure. ::

        class MyNewBackendBase(BackendBase):
            def __init__(self):  # is it needed?
                super().__init__()

            @availabe_shape_method('My_new_shape_name')
            def my_new_shape(self, parameters):
                raise ShapeNotSupportedError

        class MyNewBackend(MyNewBackendBase):
            ...

            def my_new_shape(self, parameters):
                ...


    Actually not needed:
    Make sure, that the base methods of each pattern function
    (e.g. :meth:`fibomat.backend.backendbase.BackendBase.rect`) is called, e.g. ::

        def rect(self, ptn: Pattern):
            super().rect(pattern)
            # ... do your stuff here

    """

    def __init__(self, **kwargs):
        """
        Args:
            `**kwargs`: optional parameters for the backend given by `**kwargs` in
                        :meth:`fibomat.project.Project.export`
        """
        if kwargs:
            print('warning: unhandled kwargs in exporting backend ("{}")'.format(', '.join(kwargs.keys())))

        super().__init__()

        self._abstract_shape_methods: Dict[shapes.Shape, Callable]
        self._implemented_shape_methods: Dict[shapes.Shape, Callable]

    def process_pattern(self, ptn: Pattern) -> None:
        """
        Add a pattern to the backend. The appropriate method is determined  by the pattern's shape class automatically.

        If pattern contains a Layout, the contained shapes will be extracted automatically.

        Args:
            ptn (Pattern): pattern to be added.
            on_unknown_shape:

        Returns:
            None
        """

        def dispatch(extracted_ptn):
            try:
                method = self.implemented_shape_methods[type(extracted_ptn.dim_shape.shape)]
                return method(self, extracted_ptn)
            except KeyError:
                # try bases classes
                for base in inspect.getmro(extracted_ptn.dim_shape.shape.__class__):
                    try:
                        method = self.implemented_shape_methods[base]
                        return method(self, extracted_ptn)
                    except KeyError:
                        pass
                self.process_unknown(extracted_ptn)

        if isinstance(ptn.dim_shape, layout.LayoutBase):
            # ptn.dim_shape[0]: layout.LayoutBase
            for extracted_shape in ptn.dim_shape.layout_elements():
                if isinstance(extracted_shape, layout.LayoutBase):
                    self.process_pattern(Pattern(extracted_shape, ptn.mill, ptn.raster_style, **ptn.kwargs, description=ptn.description))
                else:
                    dispatch(
                        Pattern(extracted_shape, ptn.mill, ptn.raster_style, **ptn.kwargs, description=ptn.description)
                    )
        else:
            dispatch(ptn)

    def process_unknown(self, ptn: Pattern) -> None:
        """
        Process a shape, which do not have a registered shape handler.
        Raises an exception by default.

        Args:
            ptn (Pattern):  pattern to be added.

        Returns:
            None
        """
        raise ShapeNotSupportedError(f'Shape type = {ptn.dim_shape.shape.__class__}')

    # @staticmethod
    # def _check_pattern_shape_type(expected_shape_type_: typing.Type[Shape], patter: Pattern):
    #     if expected_shape_type_.__class__ != patter.shape.__class__:
    #         raise RuntimeError('Expected shape of class {} but got {}'.format(
    #             expected_shape_type_.__class__.__name__, patter.shape.__class__.__name__
    #         ))

    @property
    def shape_methods(self) -> Dict[shapes.Shape, Callable]:
        """Dict[Shape, str]: all (abstract) shape methods"""
        return self._abstract_shape_methods

    @property
    def implemented_shape_methods(self) -> Dict[shapes.Shape, Callable]:
        """Dict[Shape, str]: implemented shape methods of backend"""
        return self._implemented_shape_methods

    def process_site(self, new_site: Site) -> None:
        """
        Adds a :class:`fibomat.site.Site` to the backend. Note, that this method processes all patterns contained in
        the site.
        Use the following example as reference, if a backend overwrites this method. ::

            ...
            def process_site(self, site: Site):
                # ...
                # do the backend specific stuff here (e.g. initialize a new patterning site)

                # call base method to add all patterns automatically
                super().process_site(site)
            ...

        Args:
            new_site (Site): site to be added.

        Returns:
            None
        """
        for ptn in new_site.patterns:
            self.process_pattern(ptn)

    def save(self, filename: PathLike) -> None:
        """
        Saves the exported project to file.
        Args:
            filename (PathLike): filename

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.Spot)
    def spot(self, ptn: Pattern[shapes.Spot]) -> None:
        """
        Adds pattern with `Spot` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `Spot` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.Line)
    def line(self, ptn: Pattern[shapes.Line]) -> None:
        """
        Adds pattern with `Line` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `Line` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.Rect)
    def rect(self, ptn: Pattern[shapes.Rect]) -> None:
        """
        Adds pattern with `Rect` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `Rect` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.Ellipse)
    def ellipse(self, ptn: Pattern[shapes.Ellipse]) -> None:
        """
        Adds pattern with `Ellipse` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `Ellipse` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.Circle)
    def circle(self, ptn: Pattern[shapes.Circle]) -> None:
        """
        Adds pattern with `Ellipse` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `Ellipse` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.Arc)
    def arc(self, ptn: Pattern[shapes.Arc]) -> None:
        """
        Adds pattern with `Arc` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `Arc` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.ArcSpline)
    def arc_spline(self, ptn: Pattern[shapes.ArcSpline]) -> None:
        """
        Adds pattern with `ArcSpline` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `ArcSpline` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.ParametricCurve)
    def parametric_curve(self, ptn: Pattern[shapes.ParametricCurve]) -> None:
        """
        Adds pattern with `Curve` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `Curve` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.RasterizedPoints)
    def rasterized_points(self, ptn: Pattern[shapes.RasterizedPoints]):
        """
        Adds pattern with `RasterizedPoints` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `RasterizedPoints` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(RasterizedPattern)
    def rasterized_pattern(self, ptn: Pattern[RasterizedPattern]):
        """
        Adds pattern with `RasterizedPattern` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `RasterizedPattern` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.Polyline)
    def polyline(self, ptn: Pattern[shapes.Polyline]) -> None:
        """
        Adds pattern with `Ellipse` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `Ellipse` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.Polygon)
    def polygon(self, ptn: Pattern[shapes.Polygon]) -> None:
        """
        Adds pattern with `Ellipse` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `Ellipse` as shape

        Returns:
            None
        """
        raise NotImplementedError

    @shape_type(shapes.HollowArcSpline)
    def hollow_arc_spline(self, ptn: Pattern[shapes.HollowArcSpline]) -> None:
        """
        Adds pattern with `HollowArcSpline` as shape to the backend.

        Args:
            ptn (Pattern): pattern with `HollowArcSpline` as shape

        Returns:
            None
        """
        raise NotImplementedError



"""Provide the abstract class :class:`IonBeam` and a implementation :class:`GaussBeam`."""
from typing import Union, overload, Tuple, List
import abc

import numpy as np

from fibomat.units import UnitType, QuantityType, U_, scale_factor
from fibomat.linalg import VectorLike, Vector, VectorValueError


class IonBeam(abc.ABC):
    """Base class to model an ion beam."""

    def __init__(self, beam_cutoff: int):
        """
        Args:
            beam_cutoff (int): beam_cutoff * std is used as cut-off length for flux calculations.

        Raises:
            ValueError: Raised if beam_cutoff < 1.
        """
        beam_cutoff = int(beam_cutoff)

        if beam_cutoff < 1:
            raise ValueError('beam_cutoff must not be smaller than 1.')

        self._beam_cutoff = beam_cutoff

    @abc.abstractmethod
    def __repr__(self):
        pass

    @property
    @abc.abstractmethod
    def std(self) -> QuantityType:
        """Standard deviation of ion beam distribution.

        Access:
            get

        Returns:
            QuantityType
        """

    @overload
    def flux_at(
        self,
        position: VectorLike, beam_maxima: VectorLike,
        position_unit: UnitType
    ) -> QuantityType: ...

    @overload
    def flux_at(
        self,
        position: VectorLike, beam_maxima: List[VectorLike],
        position_unit: UnitType
    ) -> Tuple[np.ndarray, QuantityType]: ...

    @abc.abstractmethod
    def flux_at(
        self,
        position: VectorLike, beam_maxima: Union[List[VectorLike], VectorLike],
        position_unit: UnitType
    ) -> Union[Tuple[np.ndarray, UnitType], QuantityType]:
        """Calculate the ion flux at `position` if the ion beam maximus is at beam_maximum.

        Args:
            position (VectorLike): position at which the flux is calculated.
            beam_maxima (VectorArrayLike, VectorLike): position of beam maximum/maxima
            position_unit (UnitType): length unit of positions and beam_maximum

        Returns:
            Union[QuantityType, Tuple[np.ndarray, QuantityType]]:
                QuantityType is returned if position is VectorLike and Tuple[np.ndarray, QuantityType] if positions is
                VectorArrayLike.
        """

    def nominal_flux_per_spot(self) -> QuantityType:
        """Calculate the ion flux of a single, isolated spot.

        Returns:
            QuantityType
        """
        flux, unit = self.flux_at((0, 0), (0, 0), U_('µm'))
        return flux * unit

    def nominal_flux_per_spot_on_line(self, pitch: QuantityType) -> QuantityType:
        """Calculate the ion flux of a spot on line with pitch `pitch`.

        Args:
            pitch (QuantityType): pitch on line

        Returns:
            QuantityType
        """
        pitch = pitch.to('µm')

        n_considered_spots = int((self.std * self._beam_cutoff / pitch).to_reduced_units().m)

        points = np.c_[
            pitch.m * np.arange(-n_considered_spots, n_considered_spots + 1),
            np.zeros(2*n_considered_spots + 1)
        ]

        fluxes, flux_unit = self.flux_at((0, 0), points, pitch.units)

        return np.sum(fluxes) * flux_unit

    def nominal_flux_per_spot_in_rect(self, pitch_x: QuantityType, pitch_y: QuantityType) -> QuantityType:
        """Calculate the ion flux of a spot on rectangular grid with pitches `pitch_x` and `pitch_y`.

        Args:
            pitch_x (QuantityType): pitch in x direction
            pitch_y (QuantityType): pitch in y direction

        Returns:
            QuantityType
        """
        pitch_x = pitch_x.to('µm')
        pitch_y = pitch_y.to('µm')

        n_considered_spots_x = int((self.std * self._beam_cutoff / pitch_x).to_reduced_units().m) + 1
        n_considered_spots_y = int((self.std * self._beam_cutoff / pitch_y).to_reduced_units().m) + 1

        x_grid = np.arange(-n_considered_spots_x, n_considered_spots_x+1) * pitch_x.m
        y_grid = np.arange(-n_considered_spots_y, n_considered_spots_y+1) * pitch_y.m

        points = np.dstack(np.meshgrid(x_grid, y_grid)).reshape(-1, 2)
        points = points[points[:, 0]**2 + points[:, 1]**2 < (self.std * self._beam_cutoff).to('µm').m**2]

        fluxes, flux_unit = self.flux_at((0, 0), points, U_('µm'))

        return np.sum(fluxes) * flux_unit


class GaussBeam(IonBeam):
    """Gauss function model of an ion beam."""

    def __init__(self, fwhm: QuantityType, current: QuantityType, beam_cutoff: int = 10):
        """

        Args:
            fwhm (QuantityType): full width half maximum of the beam profile.
            current (QuantityType): total beam current
            beam_cutoff (int, optional): beam cut-off, default to 10.
        """
        super().__init__(beam_cutoff)

        self._fwhm = fwhm
        self._current = current

        self._std = fwhm.to('µm') / (2 * np.sqrt(2 * np.log(2)))

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}'
            f'(fwhm={self._fwhm!r}, current={self._current!r}, beam_cutoff={self._beam_cutoff})'
        )

    @property
    def std(self) -> QuantityType:
        return self._std

    def flux_at(
        self,
        position: VectorLike, beam_maxima: Union[List[VectorLike], VectorLike],
        position_unit: UnitType
    ) -> Union[Tuple[np.ndarray, UnitType], QuantityType]:
        # pylint: disable=invalid-name
        try:
            beam_maxima = Vector(beam_maxima)  # type: ignore
        except VectorValueError:
            try:
                beam_maxima = np.asarray([Vector(vec) for vec in beam_maxima])
            except VectorValueError:
                raise ValueError('I do not understand beam_maxima\' type.')  # pylint: disable=raise-missing-from

        position = Vector(position) * scale_factor(U_('µm'), position_unit)

        beam_maxima *= scale_factor(U_('µm'), position_unit)

        # beam_maxima = np.asarray(beam_maxima)

        if isinstance(beam_maxima, Vector):
            x = position.x - beam_maxima[0]
            y = position.y - beam_maxima[1]
        else:
            x = position.x - beam_maxima[:, 0]  # type: ignore
            y = position.y - beam_maxima[:, 1]  # type: ignore

        two_sigma_squared = 2 * (self._std * self._std).m

        # sqrt(2 * np.pi) ?
        amplitude = self._current.to('pA') / (2 * np.pi * self._std * self._std)

        fluxes = np.exp(-(x * x + y * y) / two_sigma_squared)

        if isinstance(beam_maxima, Vector):
            return fluxes * amplitude

        return fluxes * amplitude.m, amplitude.units



from typing import Tuple

from fibomat.rasterizing.styles import areastyle
from fibomat.shapes import rasterizedpoints
from fibomat import shapes
from fibomat import units


class Spiral(areastyle.AreaStyle):
    def __init__(self, pitch: units.LengthQuantity, spiral_pitch: units.LengthQuantity):
        self._pitch = pitch
        self._spiral_pitch = spiral_pitch

    def rasterize(
            self,
            dim_shape: Tuple[shapes.Shape, units.LengthUnit],
            repeats: int,
            time_unit: units.TimeQuantity,
            length_unit: units.LengthUnit
    ) -> rasterizedpoints.RasterizedPoints:
        pass

    def explode(self, *args):
        pass



"""Provide the :class:`Group` class."""
# pylint: disable=protected-access
from __future__ import annotations
from typing import Optional, List, TypeVar

from fibomat.layout.groups.group_base import GroupBase
from fibomat.linalg import Vector, Transformable, BoundingBox, VectorLike
from fibomat.units import U_


class Group(GroupBase[Transformable, Vector, BoundingBox], Transformable):
    def __init__(
        self,
        elements: List[Transformable],
        description: Optional[str] = None
    ):
        super().__init__(elements=elements, description=description)

    def __mul__(self, other):
        if isinstance(other, U_):
            from fibomat.layout.groups.dim_group import DimGroup
            return DimGroup([elem * other for elem in self.elements], description=self.description)
        raise NotImplementedError




from bokeh.models import ColumnDataSource, Tool, Inspection

from bokeh.core.properties import String, DashPattern, Color, Float
from bokeh.util import compiler as bokeh_comp


class MeasureTool(Inspection):
    """
    A measure tool for bokeh plots!
    """

    # https://github.com/bokeh/bokeh/issues/9412
    __view_module__ = "bokeh"

    measure_unit = String(default='', help='')
    line_dash = DashPattern(default='solid', help='')
    line_color = Color(default="black", help='')
    line_width = Float(default=1, help='')
    line_alpha = Float(default=1.0, help='')



"""Provide the offsetting functionality."""
from typing import List, Optional

from fibomat import _libfibomat
from fibomat.shapes import ArcSpline


def offset(curve: ArcSpline, delta: float, direction: Optional[str] = None) -> List[ArcSpline]:
    """Offset a list of curves. The offset direction depends on the orientation of the curves and the sign of delta.

    ::

                         | delta > 0 | delta < 0 |
        -----------------+-----------+-----------+
        pos. orientation | inwards   | outwards  |
        -----------------+-----------+-----------+
        neg. orientation | outwards  | inwards   |
        -----------------+-----------+-----------+

    Args:
        curve (ArcSpline): curve to be offsetted.
        delta (float): offset distance
        direction (str, optional): offset direction. Only needed if curve is closed. Should be `inwards` or `outwards`.

    Returns:
        List[ArcSplines]

    Raises:
        ValueError: Raised if direction is not provided an curve is closed.
    """
    # if delta <= 0.:
    #     raise ValueError("delta <= 0.")

    raw_curve = curve.arc_spline_impl

    if curve.is_closed:

        if not direction:
            raise ValueError('direction must be provided if curve is closed.')

        if direction == 'outwards':
            if curve.orientation:
                delta *= -1
        else:
            if not curve.orientation:
                delta *= -1

    res_raw_curves = _libfibomat.offset_curve(raw_curve, delta)

    return [ArcSpline(raw_curve) for raw_curve in res_raw_curves]


def offset_with_islands(islands: List[ArcSpline], delta: float, outer_curve: Optional[ArcSpline] = None):
    """Offset a region with islands. The offset direction will always be outwards.

    Args:
        islands (List[ArcSpline]): list of islands. All islands must be closed curves and orientation must be True.
        delta (float): offset distance
        outer_curve (ArcSpline, optional): optional outer curve, which will be offsetted inwards.

    Returns:
        Dict[str, Any]:
            dict with keys 'islands' and 'outer_curve'. 'islands' contains a list of new islands and 'outer_curve'
            contains a new outer curve if outer_curve was not None.
    """
    # if delta <= 0.:
    #     raise ValueError("delta <= 0.")

    raw_islands = [island.arc_spline_impl for island in islands]
    raw_outer_curve = outer_curve.arc_spline_impl if outer_curve else None

    res_raw_curves = _libfibomat.offset_with_islands(raw_islands, raw_outer_curve, delta)

    islands = [ArcSpline(raw_curve) for raw_curve in res_raw_curves[0]]
    outer_curve = [ArcSpline(raw_curve) for raw_curve in res_raw_curves[1]]

    return {'islands': islands, 'outer_curve': outer_curve}


SAFETY: int = 1000
"""max steps for deflate"""


def deflate(
    arc_spline: ArcSpline, pitch: float, *, n_steps: Optional[int] = None, distance: Optional[float] = None
):
    """Deflate a given arc spline (completely).

    The original curve is not included.

    Args:
        arc_spline (ArcSpline): curve to be inflated
        pitch (float): offset delta.
        n_steps (int, optional): if provided at maximum n_steps will be performed.
        distance (float, optional): if provided, a total distance 'distance' will be deflated.

    Returns:
        List[ArcSpline]

    Raises:
        ValueError: Raised if steps < 0.
        ValueError: Raised if distance < 0.
        ValueError: Raised if steps and distance are provided.
        RuntimeError: Raised if more than :attr:`SAFETY` steps are performed.
    """
    if n_steps and not distance:
        if n_steps < 0:
            raise ValueError('steps must be greater than 0.')
        max_steps = n_steps
    elif not n_steps and distance:
        if distance < 0.:
            raise ValueError('distance must be greater than 0.')
        max_steps = int(distance / pitch)
    else:
        max_steps = SAFETY
        if n_steps and distance:
            raise ValueError('distance and steps cannot be set both.')

    deflated_curves = []

    deflate_stack = [arc_spline]
    new_deflate_stack = []

    finished = False

    steps = 0

    while not finished and steps < max_steps:
        for to_be_inflated in deflate_stack:
            newly_deflated = offset(to_be_inflated, pitch, 'inwards')
            if newly_deflated:
                deflated_curves.extend(newly_deflated)
                new_deflate_stack.extend(newly_deflated)

        deflate_stack = new_deflate_stack
        new_deflate_stack = []

        steps += 1
        if steps == SAFETY:
            raise RuntimeError('steps > SAFETY. If you know you need so many steps, increase SAFETY.')

        if not deflate_stack:
            finished = True

    return deflated_curves


def inflate(
    arc_spline: ArcSpline, pitch: float, *, n_steps: Optional[int] = None, distance: Optional[float] = None
):
    """Inflate a given arc spline (completely).

    The original curve is not included.

    Note, any one of n_steps and distance must be provided.

    Args:
        arc_spline (ArcSpline): curve to be inflated
        pitch (float): offset delta.
        n_steps (int, optional): if provided at maximum n_steps will be performed.
        distance (float, optional): if provided, a total distance 'distance' will be deflated.

    Returns:
        List[ArcSpline]

    Raises:
        ValueError: Raised if steps < 0.
        ValueError: Raised if distance < 0.
        ValueError: Raised if none of steps and distance are provided.
    """
    if n_steps and not distance:
        if n_steps < 0:
            raise ValueError('steps must be greater than 0.')
        max_steps = n_steps
    elif not n_steps and distance:
        if distance < 0.:
            raise ValueError('distance must be greater than 0.')
        max_steps = int(distance / pitch)
    else:
        raise ValueError('Any of distance or steps must be set.')

    inflated_curves = []
    inflate_stack = [arc_spline]
    new_inflate_stack = []

    for _ in range(max_steps):
        for to_be_inflated in inflate_stack:
            newly_deflated = offset(to_be_inflated, pitch, 'outwards')
            if newly_deflated:
                inflated_curves.extend(newly_deflated)
                new_inflate_stack.extend(newly_deflated)

        inflate_stack = new_inflate_stack
        new_inflate_stack = []

    return inflated_curves



"""Provides the functions :func:`translate`, :func:`rotate`, :func:`scale` and :func:`mirror` to build chained
transformations.
"""
from __future__ import annotations

from typing import List, Optional, Union, Generic, TypeVar

from fibomat.linalg.vectors import VectorLike, Vector, DimVector

VectorT = TypeVar('VectorT', Vector, DimVector)


class _TransformationBuilder(Generic[VectorT]):  # pylint: disable=too-few-public-methods
    def __init__(self):
        self._transformations: List[_TransformationBuilder[VectorT]] = [self]

    @property
    def transformations(self) -> List[_TransformationBuilder[VectorT]]:
        """Saved transformations.

        Access:
            get

        Returns:
            List[_TransformationBuilder]
        """
        return self._transformations

    def __or__(self, other: _TransformationBuilder[VectorT]) -> _TransformationBuilder[VectorT]:
        if not isinstance(other, _TransformationBuilder):
            raise TypeError('other is not an TransformationBuilder object.')

        self._transformations.append(other)
        return self


class _TranslationBuilder(_TransformationBuilder[VectorT]):  # pylint: disable=too-few-public-methods
    def __init__(self, trans_vec: VectorT):
        super().__init__()
        self.trans_vec = trans_vec


class _RotationBuilder(_TransformationBuilder[VectorT]):  # pylint: disable=too-few-public-methods
    def __init__(self, theta: float, origin: Optional[Union[VectorT, str]] = None):
        super().__init__()
        self.theta = theta
        self.origin = origin


class _ScaleBuilder(_TransformationBuilder[VectorT]):  # pylint: disable=too-few-public-methods
    def __init__(self, fac: float, origin: Optional[Union[VectorT, str]] = None):
        super().__init__()
        self.fac = fac
        self.origin = origin


class _MirrorBuilder(_TransformationBuilder[VectorT]):  # pylint: disable=too-few-public-methods
    def __init__(self, mirror_plane: VectorT):
        super().__init__()
        self.mirror_plane = mirror_plane


def translate(trans_vec: VectorT):
    """Translate a object by `trans_vec`.

    Args:
        trans_vec (VectorLike): translation vector

    Returns:
        _TranslationBuilder
    """
    return _TranslationBuilder(trans_vec)


def rotate(theta: float, origin: Optional[Union[VectorT, str]] = None):
    """Rotate a object `origin` with angle `theta` in math. positive direction (counterclockwise).

    Args:
        theta (float): rotation angle in rad
        origin (Vector, str, optional):
            origin of rotation. If not set, (0,0) is used as origin. If origin == 'center', the
            :attr:`Transformable.center` of the object will be used. The same applies for the case that
            origin == 'origin' with the :attr:`Transformable.origin` property. Default to None.

    Returns:
        _RotationBuilder
    """
    return _RotationBuilder(theta, origin)


def scale(fac: float, origin: Optional[Union[VectorT, str]] = None):
    """Scale a object homogeneously about `origin` with factor `fac`.

    Args:
        fac (float): rotation angle in rad
        origin (Optional[Union[linalg.VectorLike, str]], optional):
            origin of rotation. If not set, (0,0) is used as
            origin. If origin == 'center', the
            :attr:`Transformable.center` of the object will
            be used. The same applies for the case that
            origin == 'origin' with the
            :attr:`Transformable.origin` property. Default
            to None.

    Returns:
        _ScaleBuilder
    """
    return _ScaleBuilder(fac, origin)


def mirror(mirror_plane: VectorT):
    """Mirror a object mirrored about `mirror_plane`.

    Args:
        mirror_plane (VectorLike): mirror plane to be used.

    Returns:
        Transformable
    """
    return _MirrorBuilder(mirror_plane)



from fibomat.shapes.polyline import Polyline
from typing import Optional, Dict
import io

from PIL import Image, ImageDraw

import numpy as np

import svgwrite
import svgwrite.container
import svgwrite.shapes
import svgwrite.extensions

from fibomat.backend import BackendBase
from fibomat.site import Site
from fibomat.units import U_, Q_, scale_factor
from fibomat.utils import PathLike
from fibomat.pattern import Pattern
from fibomat.shapes import Rect, Polygon, Circle
from fibomat.linalg import DimBoundingBox, scale, translate, Vector, DimVector


class BitmapBackend(BackendBase):
    def __init__(self, pixel_size: Q_ = Q_('2 nm'), description: Optional[str] = None):
        super().__init__()

        self._description = description
        self._pixel_size = pixel_size

        self._site = None
        # self._current_layer: Optional[Dict] = None
        self._center: Optional[Vector] = None

        self._image: Optional = None

        self._total_bounding_box = None

    def save(self, filename: str):
        if self._image:
            self._image.save(filename)
        else:
            raise RuntimeError('No site added.')

    def image(self):
        return self._image

    def process_site(self, new_site: Site) -> None:
        if self._image:
            raise RuntimeError('Only one site is allowed in bitmap backend.')

        self._site = new_site

        self._shift = DimVector(new_site.fov[0] / 2, new_site.fov[1] / 2)

        width = int(np.rint((new_site.fov[0] / self._pixel_size).to_base_units()))
        height = int(np.rint((new_site.fov[1] / self._pixel_size).to_base_units()))
        self._image = Image.new("L", (width, height), 'black')

        super().process_site(new_site)

    def _to_pixel(self, value):
        return int(np.rint((value / self._pixel_size).to_base_units()))

    # def _pixel_scale_factor(self, other_unit):
    #     return scale_factor(self._pixel_scale, other_unit) / self._pixel_scale.m

    # def _scale_and_shift_shape(self, ptn: Pattern):
    #     fak = self._pixel_scale_factor(ptn.dim_shape.unit)

    #     return ptn.dim_shape.shape.transformed(scale(fak) | translate(self._current_layer_center))

    def rect(self, ptn: Pattern[Rect]) -> None:
        # return super().rect(ptn)
        ptn = ptn.translated(self._shift)

        # print(self._image.size)

        points = ptn.dim_shape.shape.to_arc_spline().vertices[:, :2]
        pixel_points = [(self._to_pixel(x * ptn.dim_shape.unit), self._image.size[1] -self._to_pixel(y * ptn.dim_shape.unit))  for x, y in points]

        # print(pixel_points)

        draw = ImageDraw.Draw(self._image)
        draw.polygon(pixel_points, 'white')


    # def polygon(self, ptn: Pattern[Polygon]) -> None:
    #     return super().polygon(ptn)


"""Provides the :class:`Circle` class."""
# pylint: disable=invalid-name
from __future__ import annotations
from typing import Optional

import numpy as np

from fibomat.shapes.shape import Shape
from fibomat.linalg import Vector, VectorLike, BoundingBox
from fibomat.shapes.arc_spline import ArcSpline, ArcSplineCompatible


class Circle(Shape, ArcSplineCompatible):
    """2-dim circle."""

    def __init__(self, r: float, center: Optional[VectorLike] = None, description: Optional[str] = None):
        """

        Args:
            r (float): radius
            center (:class:`~fibomat.linalg.vectortypes.VectorLike`, optional): center of circle, default to (0, 0)
            description (str, optional): description

        Raises:
            ValueError: Raised if r <= 0.
        """
        super().__init__(description)

        self._center = Vector(center) if center is not None else Vector()
        self._r = float(r)
        if self._r <= 0:
            raise ValueError('radius <= 0.')

    @classmethod
    def from_points(cls, p1: VectorLike, p2: VectorLike, p3: VectorLike, description: Optional[str] = None):
        """Create circumscribed circle (from three points).

        Args:
            p1 (VectorLike): first points
            p2 (VectorLike): second points
            p3 (VectorLike): third points
            description (str, optional): description

        Returns:
            Circle
        """
        # https://en.wikipedia.org/wiki/Circumscribed_circle#Cartesian_coordinates_2
        a = Vector(p1)  # A
        b = Vector(p2) - a  # B
        c = Vector(p3) - a  # C

        inv_d = 1 / (2 * (b.x * c.y - b.y * c.x))

        b2 = b.x*b.x + b.y*b.y
        c2 = c.x*c.x + c.y*c.y

        u = Vector(inv_d * (c.y * b2 - b.y * c2), inv_d * (b.x * c2 - c.x * b2))
        r = np.sqrt(u.x*u.x + u.y*u.y)

        return cls(r, u + a, description)

    def __repr__(self) -> str:
        return '{}(r={!r}, center={!r})'.format(
            self.__class__.__name__, self.r, self.center)

    def to_arc_spline(self) -> ArcSpline:
        return ArcSpline(
            [(*(self._center + Vector(self._r, 0.)), 1.), (*(self._center - Vector(self._r, 0.)), 1.)],
            True,
            self.description
        )

    @property
    def r(self) -> float:
        """Radius.

        Access:
            get

        Returns:
            float
        """
        return self._r

    @property
    def center(self) -> Vector:
        return self._center

    @property
    def bounding_box(self) -> BoundingBox:
        return BoundingBox(self._center-(self._r, self._r), self._center + (self._r, self._r))

    @property
    def is_closed(self) -> bool:
        return True

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._center += Vector(trans_vec)

    def _impl_rotate(self, theta: float) -> None:
        self._center = self._center.rotated(theta)

    def _impl_scale(self, fac: float) -> None:
        self._center *= float(fac)
        self._r *= float(fac)

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        self._center = self._center.mirrored(mirror_axis)



import sys
import os
import pathlib
import shutil

from setuptools import find_packages, Extension

try:
    from skbuild import setup
    from skbuild.command.build_ext import build_ext
except ImportError:
    print('Please update pip, you need pip 10 or greater,\n'
          ' or you need to install the PEP 518 requirements in pyproject.toml yourself', file=sys.stderr)
    raise


class BokehExtension(Extension):
    def __init__(self, name, ext_dir):
        Extension.__init__(self, name, sources=[])
        self.ext_dir = os.path.abspath(ext_dir)


class BuildBokehExt(build_ext):
    def run(self):
        from bokeh.ext import build

        for ext in self.extensions:
            if not build(ext.ext_dir, rebuild=True):
                raise RuntimeError('Could not build bokeh extension.')

            out_path = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
            shutil.copy2(
                f'{ext.ext_dir}/dist/bokeh-measuretool.min.js',
                f'{out_path}/fibomat/default_backends/bokeh-measuretool.min.js'
            )


here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')


setup(
    name="fibomat",
    version="0.3.9",
    description="fib-o-mat is a toolbox to generate patterns for focused ion beam instruments.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Victor Deinhart',
    author_email='victor.deinhart@helmholtz-berlin.de',
    url='https://gitlab.com/viggge/fib-o-mat',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: C++',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
    ],
    keywords='focused ion beam, fib, pattern, patterning, beam path generation,',
    license="GPLv3",
    packages=find_packages(exclude=['test*', 'custom_backends*']),
    python_requires='>=3.8',
    package_data={'fibomat': ['py.typed', 'default_backends/measure_band.ts']},
    cmake_install_dir='fibomat',
    ext_modules=[BokehExtension('bokeh-measuretool', 'bokeh-measuretool')],
    cmdclass={'build_ext': BuildBokehExt},
    install_requires=[
        'numpy', 'scipy', 'sympy', 'bokeh', 'pint', 'ezdxf', 'numba', 'vispy', 'pyqt5', 'splipy', 'frozenlist',
        'pillow', 'svgwrite', 'pyhershey', 'jinja2'
    ],
    extras_require={
        'docs': [
            'sphinx', 'recommonmark', 'pydata_sphinx_theme', 'sphinxemoji', 'bokeh'
        ],
        'testing': [
            'pytest', 'pytest-coverage', 'pytest-mock'
        ],
        'dev': [
            'bump2version', 'twine'
        ]
    },
    entry_points={
        'console_scripts': [
            'beam_simulation = fibomat.beam_simulation:run',
        ],
    },
    project_urls={  # Optional
        'Bug Reports': 'https://gitlab.com/viggge/fib-o-mat/-/issues',
        'Source': 'https://gitlab.com/viggge/fib-o-mat/',
        'Documentation': "https://fib-o-mat.readthedocs.io/en/latest/",
    },
)



from typing import Tuple, Union, Optional, TYPE_CHECKING

import numpy as np

from fibomat.shapes.line import Line
from fibomat.shapes.arc import Arc
from fibomat.shapes.biarc import Biarc

if TYPE_CHECKING:
    from fibomat.shapes.parametric_curve import ParametricCurve

from .utils import _make_perp_vector, _Bisector, _PerpBisector, _circle_circle_intersection, _make_fixed_tangent_curvature


# def _make_biarc_segment(parametric_curve: 'ParametricCurve', biarc_domain: Tuple[float, float], epsilon: float):
#     p_0, p_1 = parametric_curve.f(biarc_domain)
#
#     t_0, k_0 = _make_fixed_tangent_curvature(parametric_curve, biarc_domain[0], epsilon, 'left')
#     t_1, k_1 = _make_fixed_tangent_curvature(parametric_curve, biarc_domain[1], epsilon, 'right')
#
#     t_0 /= np.linalg.norm(t_0)
#     t_1 /= np.linalg.norm(t_1)
#
#     t_0_perp = _make_perp_vector(t_0)
#     t_1_perp = _make_perp_vector(t_1)
#
#     b_0 = _PerpBisector(p_0, p_1)
#     b_1 = _PerpBisector(p_0 + t_0, p_1 + t_1)
#
#     if b_0.parallel_to(b_1):
#         return Line(p_0, p_1),
#     else:
#         c_j = b_0.intersect_at(b_1)
#
#         # calc J
#         arc_j_1 = Arc.from_points_center(p_0, p_1, c_j, True)
#         arc_j_2 = Arc.from_points_center(p_0, p_1, c_j, False)
#
#         if arc_j_1.length < arc_j_2.length:
#             j = arc_j_1.midpoint
#             sweep_dir = True
#         else:
#             j = arc_j_2.midpoint
#             sweep_dir = False
#
#         c_0 = _PerpBisector(p_0, j).intersect_at(_Bisector(p_0, p_0 + t_0_perp))
#         arc_0 = Arc.from_points_center_tangent(p_0, j, c_0, unit_tangent_start=t_0)
#
#         c_1 = _PerpBisector(j, p_1).intersect_at(_Bisector(p_1, p_1 + t_1_perp))
#         arc_1 = Arc.from_points_center_tangent(j, p_1, c_1, unit_tangent_end=t_1)
#
#         return arc_0, arc_1


def _make_biarc_from_spiral_segment(
    parametric_curve: 'ParametricCurve', biarc_domain: Tuple[float, float], epsilon: float
):
    # u_0, u_1 = biarc_domain
    p_0, p_1 = parametric_curve.f(biarc_domain)

    t_0, k_0 = _make_fixed_tangent_curvature(parametric_curve, biarc_domain[0], epsilon, 'left')
    t_1, k_1 = _make_fixed_tangent_curvature(parametric_curve, biarc_domain[1], epsilon, 'right')
    t_0 /= np.linalg.norm(t_0)
    t_1 /= np.linalg.norm(t_1)

    # k_0, k_1 = parametric_curve.curvature(biarc_domain)

    b_0 = _PerpBisector(p_0, p_1)
    b_1 = _PerpBisector(p_0 + t_0, p_1 + t_1)

    if b_0.parallel_to(b_1):
        return Line(p_0, p_1),
    else:
        c_j = b_0.intersect_at(b_1)
        r_j = np.linalg.norm(c_j - p_0)

        if np.isclose(k_0, k_1):
            arc = Arc.from_points_center_tangent(start=p_0, end=p_1, center=c_j, unit_tangent_start=t_0)
            return arc.split_at(arc.theta/2)

        if abs(k_0) > abs(k_1):
            r_alpha = abs(1. / k_0)
            p_alpha = p_0
            t_alpha = t_0
            t_alpha_perp = np.sign(k_0) * _make_perp_vector(t_alpha)
            c_alpha = p_alpha + r_alpha * t_alpha_perp
        else:
            r_alpha = abs(1. / k_1)
            p_alpha = p_1
            t_alpha = t_1
            t_alpha_perp = np.sign(k_1) * _make_perp_vector(t_alpha)
            c_alpha = p_alpha + r_alpha * t_alpha_perp

        intersections_j_alpha = _circle_circle_intersection((c_j, r_j), (c_alpha, r_alpha))

        only_other_arc = False

        if len(intersections_j_alpha) == 0:
            # points lie on a circle !?
            # this should not happen if curvatures are not identical of start end endpoint
            raise RuntimeError
            # p_mean = p_0 + p_1
            # j = p_mean / np.linalg.norm(p_mean) * r_j
        elif len(intersections_j_alpha) == 2:
            if np.allclose(intersections_j_alpha[0], p_alpha):
                j = intersections_j_alpha[1]
            elif np.allclose(intersections_j_alpha[1], p_alpha):
                j = intersections_j_alpha[0]
            else:
                raise RuntimeError
        else:
            if not np.allclose(intersections_j_alpha[0], p_alpha):
                raise RuntimeError
            j = intersections_j_alpha[0]
            only_other_arc = True

        if abs(k_0) > abs(k_1):
            p_beta = p_1
            t_beta = t_1
            t_beta_perp = _make_perp_vector(t_beta)
            c_beta = _Bisector(j, c_alpha).intersect_at(_Bisector(p_beta, p_beta + t_beta_perp))

            arc_beta = Arc.from_points_center_tangent(j, p_beta, c_beta, unit_tangent_end=t_beta)

            if only_other_arc:
                return arc_beta,

            arc_alpha = Arc.from_points_center_tangent(p_alpha, j, c_alpha, unit_tangent_start=t_alpha)

        else:
            p_beta = p_0
            t_beta = t_0
            t_beta_perp = _make_perp_vector(t_beta)
            c_beta = _Bisector(j, c_alpha).intersect_at(_Bisector(p_beta, p_beta + t_beta_perp)) # plus or minus?

            arc_beta = Arc.from_points_center_tangent(p_beta, j, c_beta, unit_tangent_start=t_beta)

            if only_other_arc:
                return arc_beta,

            arc_alpha = Arc.from_points_center_tangent(j, p_alpha, c_alpha, unit_tangent_end=t_alpha)

        if abs(k_0) > abs(k_1):
            return arc_alpha, arc_beta
        else:
            return arc_beta, arc_alpha


def _fitting_error(
    param_curve: 'ParametricCurve',
    seg: Union[Tuple[Arc, Arc], Line],
    rasterized_curve_points: np.ndarray
):
    def closest_point(points_, single):
        delta = points_ - single
        dist_ = np.einsum('ij,ij->i', delta, delta)
        return np.argmin(dist_)

    # i_start = np.searchsorted(rasterized_curve_points[:, 0], biarc_domain[0], side='left')
    # i_end = np.searchsorted(rasterized_curve_points[:, 0], biarc_domain[1], side='right')

    if len(seg) == 2:
        biarc = seg

        if len(rasterized_curve_points) == 2:
            # biarcs interpolate points
            return 0.

        j_arcs = biarc[0].end
        # normal = _Bisector(j_arcs, biarc[0].center)
        # res = optimize.root(lambda u: normal(u[0]) - param_curve.f(u[1]), np.array([0., (biarc_domain[1]-biarc_domain[0])/2]))
        # that's a crude approximation to the above lines
        i_j = closest_point(rasterized_curve_points[:, 1:], j_arcs)

        if i_j == 0:
            # biarc[1][i_j] == start of biarc[0] => shift i_j to next point
            i_j += 1

        points_arc_1 = rasterized_curve_points[:i_j, 1:] - biarc[0].center
        dist_arc_1 = np.abs(np.sqrt(np.einsum('ij,ij->i', points_arc_1, points_arc_1)) - biarc[0].radius)

        points_arc_2 = rasterized_curve_points[i_j:, 1:] - biarc[1].center
        dist_arc_2 = np.abs(np.sqrt(np.einsum('ij,ij->i', points_arc_2, points_arc_2)) - biarc[1].radius)

        from fibomat import Sample, U_
        from fibomat.shapes import Spot

        # s = Sample()
        # s.add_annotation(biarc[0] * U_('µm'), color='green')
        # s.add_annotation(biarc[1] * U_('µm'), color='red')
        #
        # for p in points_arc_1:
        #     s.add_annotation(Spot(p + biarc[0].center) * U_('µm'), color='green')
        #
        # for p in points_arc_2:
        #     s.add_annotation(Spot(p + biarc[1].center) * U_('µm'), color='red')
        #
        # # for p in param_curve.rasterize_at(0.01):
        # #     s.add_annotation(Spot(p) * U_('µm'))
        #
        # s.plot()

        if len(dist_arc_1) == 0:
            return np.max(dist_arc_2)
        elif len(dist_arc_2) == 0:
            return np.max(dist_arc_1)
        else:
            return max(np.max(dist_arc_1), np.max(dist_arc_2))
    elif len(seg) == 1:
        line_: Line = seg[0]

        points = rasterized_curve_points[:, 1:]

        # https://en.wikibooks.org/wiki/Linear_Algebra/Orthogonal_Projection_Onto_a_Line
        s = np.asarray(line_.end - line_.start)
        v = points

        proj = np.outer(np.dot(v, s) / np.dot(s, s), s)

        delta = proj - v
        dist = np.einsum('ij,ij->i', delta, delta)

        return np.sqrt(np.max(dist))

    else:
        raise RuntimeError


def _approximate_param_curve_greedy(
    param_func: 'ParametricCurve',
    parameter_values: np.ndarray,
    epsilon: float
):
    # u_min, u_max = parameter_values[0], parameter_values[-1]
    # u_0, u_1 = u_min, u_min
    #
    # h = (u_max - u_min) / 2

    i_min = 0
    i_max = len(parameter_values) - 1

    # rasterized_curve_points = param_func.f(u_rasterized_int)
    rasterized_curve_points = np.empty(shape=(len(parameter_values), 3), dtype=float)
    rasterized_curve_points[:, 0] = parameter_values
    rasterized_curve_points[:, 1:] = param_func.f(parameter_values)

    from fibomat import Sample, U_
    from fibomat.shapes import Spot
    s = Sample()
    for point in rasterized_curve_points:
        s.add_annotation(Spot(point[1:]) * U_('µm'))

    segments = []

    end_reached = False
    testing = False
    step_size_reduced = False

    prev_segment = None
    prev_h: Optional[float] = None

    i_seg_start, i_seg_end = i_min, i_max
    h = (i_seg_end - i_seg_start) // 2 + 2

    if h == 0:
        raise RuntimeError

    while True:
        i_seg_end = i_seg_start + h

        if i_seg_end > i_max:
            h = i_max - i_seg_start
            # u_1 = u_max
            end_reached = True

            interval = (rasterized_curve_points[i_seg_start, 0], rasterized_curve_points[i_max, 0])
        else:
            interval = (rasterized_curve_points[i_seg_start, 0], rasterized_curve_points[i_seg_end, 0])

        # new_segment = _make_biarc_from_spiral_segment(param_func, interval)

        if h <= 2:
            p_0, p_1 = param_func.f(interval)

            t_0, _ = _make_fixed_tangent_curvature(param_func, interval[0], epsilon, 'left')
            t_1, _ = _make_fixed_tangent_curvature(param_func, interval[1], epsilon, 'right')

            new_segment = Biarc(p_0, p_1, t_0, t_1).segments
            error = 0
        else:

            # t_0, t_1 = param_func.df(interval)
            # if np.allclose(t_0, 0) or np.allclose(t_1, 0):
            #     print('dsfdsf')
            #     new_segment = _make_biarc_segment(param_func, interval, epsilon)
            # else:
            #     new_segment = _make_biarc_from_spiral_segment(param_func, interval, epsilon)
            new_segment = _make_biarc_from_spiral_segment(param_func, interval, epsilon)

            error = _fitting_error(param_func, new_segment, rasterized_curve_points[i_seg_start:i_seg_end+1])

        # print('domain', interval, i_seg_start, i_seg_end)
        # print(h, error)

        if error < epsilon:

            if end_reached:
                segments.extend(new_segment)
                break

            if testing and step_size_reduced:
                # def plot():
                #     for seg in prev_segment:
                #         s.add_annotation(seg * U_('µm'))
                #     s.plot(rasterize_pitch=Q_('0.001 µm'))
                #
                # plot()

                segments.extend(prev_segment)
                i_seg_start = i_seg_end - prev_h

                testing = False
                step_size_reduced = False
            else:
                # save current state and try again in next round with extended interval
                prev_segment = new_segment
                prev_h = h
                h *= 2

                testing = True
        else:
            end_reached = False
            if testing:
                # save state from last round
                # print('int =', (u_0, u_1 - h_temp), 'saving', 'error check')

                # def plot():
                #     for seg in prev_segment:
                #         s.add_annotation(seg * U_('µm'))
                #     s.plot(rasterize_pitch=Q_('0.001 µm'))
                #
                # plot()

                segments.extend(prev_segment)
                i_seg_start = i_seg_end - prev_h

                testing = False
            else:
                # reduce step size
                if h > 10:
                    h = h // 2
                else:
                    h -= 1

                # if h <= 3:
                #     # raise RuntimeError
                #     print('interesting')

                step_size_reduced = True

                # T0D0: Find a good criterion for aborting
                # if h < 0.01 * epsilon:
                #     raise RuntimeError('Fitting does not converge.')

    # return biarcs

    return segments



"""Provides the :class:`Describable` class."""
from typing import Optional, TypeVar
import copy


T = TypeVar('T', bound='DimTransformable')  # pylint: disable=invalid-name


class Describable:
    """This class handles optional descriptions in the fib-o-mat library."""
    def __init__(self, description: Optional[str] = None):
        """
        Args:
            description (str, optional): description
        """
       #  super().__init__()

        self._description = str(description) if description else None

    def clone(self: T) -> T:
        """Create a deepcopy of the object.

        Returns:
            Describable
        """
        return copy.deepcopy(self)

    def with_changed_description(self: T, new_descr: str) -> T:
        """Clones the object and set the description to `new_descr`.

        Args:
            new_descr: new description

        Returns:
            Describable
        """
        cloned = self.clone()
        cloned._description = str(new_descr)
        return cloned

    @property
    def description(self) -> Optional[str]:
        """Description str.

        Access:
            get

        Returns:
            Optional[str]
        """
        return self._description





from typing import Sequence, Any

import numpy as np
import pint

from fibomat.linalg.vectors.vector_base import VectorBase, FloatTypes


class Vector(VectorBase[float]):
    @staticmethod
    def _validate_type(val: Any) -> bool:
        return isinstance(val, FloatTypes)

    @staticmethod
    def _build_np_array(val: Sequence[float]) -> np.ndarray:
        return np.array(val, dtype=float)

    @staticmethod
    def _make_zero_vector():
        return [0., 0.]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __mul__(self, other: Any):
        if isinstance(other, pint.Unit):
            from .dim_vector import DimVector
            return DimVector(self._vector * other)
        else:
            return super().__mul__(other)



"""Provides the :class:`Site` class."""
from __future__ import annotations

from typing import Optional, List, Union
from fibomat.linalg.vectors.vector import Vector

import numpy as np

from fibomat.linalg import (
    DimTransformable, DimVectorLike, DimVector, DimBoundingBox
)
from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.mill import MillBase
from fibomat import layout
from fibomat.pattern import Pattern
from fibomat.shapes import DimShape


class Site(DimTransformable):
    """
    The `Site` class is used to collect shapes with its patterning settings.

    .. note:: All shape positions added to each site are interpreted relative to the site's position!

    .. note:: If fov is not passed, it will be determined by the bounding box of the added patterns. But if a fixed FOV
              is given, it is **not** checked if the added shapes fit inside the fov.

    """
    def __init__(
            self,
            dim_center: DimVectorLike,
            dim_fov: Optional[DimVectorLike] = None,
            *,
            description: Optional[str] = None
    ):
        """
        Args:
            dim_center (DimVectorLike): Center coordinate of the site.
            dim_fov (DimVectorLike, optional): The fov (field of view) to be used. If not given, the fov will be
                                               calculated automatically.
            description (str, optional): description
        """
        super().__init__(description=description)

        self._center = DimVector(dim_center)

        self._theta_vec = Vector(1, 0)

        # TODO: check if fov is valid?
        self._fov = DimVector(dim_fov) if dim_fov is not None else None

        self._patterns: List[Pattern] = []

    @property
    def fov(self) -> DimVector:
        """Field-of-view of the site.

        Access:
            get

        Returns:
            DimVector
        """
        if self._fov:
            return self._fov
        else:
            bbox = self.bounding_box
            return DimVector(bbox.width, bbox.height)

    @property
    def square_fov(self) -> DimVector:
        """Squared field-of-view of the site.

        Access:
            get

        Returns:
            DimVector
        """
        fov = self.fov
        size = fov.x if fov.x > fov.y else fov.y
        return DimVector(size, size)

    @property
    def fov_bounding_box(self) -> DimBoundingBox:
        """Bounding box given by fov and center rather from the contained patterns.

        Returns:
            DimBoundingBox
        """
        fov_x_2 = self._fov.x / 2
        fov_y_2 = self._fov.y / 2
        return DimBoundingBox(self._center - (fov_x_2, fov_x_2), self._center + (fov_x_2, fov_x_2))

    @property
    def empty(self) -> bool:
        """If True, site does not contain any shapes

        Access:
            get

        Returns:
            bool
        """
        return not self._patterns

    @property
    def bounding_box(self) -> DimBoundingBox:
        """Bounding box of the added patterns.

        Access:
            get

        Returns:
            DimBoundingBox
        """
        # bbox = DimBoundingBox(self._center, self._center)

        if not self._patterns:
            raise RuntimeError('Cannot calculate bounding box of empty site.')

        bbox = self._patterns[0].bounding_box

        for pattern in self._patterns[1:]:
            bbox = bbox.extended(pattern.bounding_box)

        return bbox

    @property
    def patterns(self):
        """Contained patterns in site

        Access:
            get

        Returns:
            List[Pattern]
        """
        return self._patterns

    @property
    def patterns_absolute(self):
        """Return a list of all patterns contained in the side which are shifted by the side's center.
        Hence, these patterns' positions are **not** relative to the site's center anymore but absolute to the
        coordinate origin.

        Access:
            get

        Returns:
            List[Pattern]
        """
        return [pattern.translated(self._center) for pattern in self._patterns]

    def create_pattern(
        self,
        dim_shape: DimShape,
        mill: MillBase,
        raster_style: RasterStyle,
        description: Optional[str] = None,
        **kwargs
    ) -> Pattern:
        """Creates a pattern in-place (returned pattern is automatically added to the site).
        The parameters are identical to the __init__method of the :class:`fibomat.pattern.Pattern` class.

        Args:
            dim_shape:
            mill:
            raster_style:
            description:
            **kwargs:

        Returns:
            Pattern
        """
        pattern = Pattern(dim_shape, mill, raster_style, description=description, **kwargs)
        self.add_pattern(pattern)
        return pattern

    def add_pattern(self, ptn: Union[Pattern, layout.LayoutBase]) -> None:
        """Adds a :class:`fibomat.pattern.Pattern` or Layoutbase[Pattern] to the site.

        Args:
            ptn (Pattern):  new pattern

        Returns:
            None
        """
        if isinstance(ptn, layout.LayoutBase):
            for extracted_pattern in ptn.layout_elements():
                self._patterns.append(extracted_pattern)
        else:
            self._patterns.append(ptn)

    def __iadd__(self, ptn:  Union[Pattern, layout.LayoutBase]) -> Site:
        """Adds a :class:`fibomat.pattern.Pattern` to the site.
        Identical to :meth:`add_pattern`

        Args:
            ptn: new pattern

        Returns:
            None
        """
        self.add_pattern(ptn)
        return self

    @property
    def center(self) -> DimVector:
        """Center of the site.

        Access:
            get

        Returns:
            DimVector
        """
        return self._center

    def _impl_translate(self, trans_vec: DimVectorLike) -> None:
        trans_vec = DimVector(trans_vec)
        self._center += DimVector(trans_vec)

    def _impl_rotate(self, theta: float, _allow_any_rot=False) -> None:
        if not _allow_any_rot:
            if not np.isclose(np.mod(theta, np.pi/2), 0.):
                raise ValueError('Sites can only be rotated by multiples of pi/2')

            if not np.isclose(np.mod(theta, np.pi), 0.):
                self._fov = DimVector(self._fov.y, self._fov.x)

        self._center = self._center.rotated(theta)

        self._theta_vec = self._theta_vec.rotated(theta)

        for ptn in self._patterns:
            ptn._impl_rotate(theta)

    def _impl_scale(self, fac: float) -> None:
        self._fov *= float(fac)
        self._center *= float(fac)

        for ptn in self._patterns:
            ptn._impl_scale(fac)

    def _impl_mirror(self, mirror_axis: DimVectorLike) -> None:
        mirror_axis = DimVector(mirror_axis)

        if not np.isclose(np.mod(mirror_axis.vector.angle_about_x_axis, np.pi/4), 0.):
            raise ValueError(
                'Sites can only be mirrored on the axes or their diagonals.'
            )

        self._center = self._center.mirrored(mirror_axis)

        self._theta_vec = self._theta_vec.mirrored(mirror_axis.vector)

        if not np.isclose(np.mod(mirror_axis.vector.angle_about_x_axis, np.pi), 0.):
            self._fov = DimVector(self._fov.y, self._fov.x)

        for ptn in self._patterns:
            ptn._impl_mirror(mirror_axis)



from typing import Tuple, Sequence

import numpy as np

from fibomat.rasterizedpattern import RasterizedPattern
from fibomat.raster_styles.rasterstyle import RasterStyle
from fibomat.units import LengthUnit, TimeUnit, scale_to, scale_factor
from fibomat.shapes import Spot, DimShape
from fibomat.mill import Mill


class SingleSpot(RasterStyle):
    def __init__(self):
        super().__init__()

    def __repr__(self) -> str:
        return '{}()'.format(self.__class__.__name__)

    @property
    def dimension(self) -> int:
        return 0

    def rasterize(
        self,
        dim_shape: DimShape,
        mill: Mill,
        out_length_unit: LengthUnit,
        out_time_unit: TimeUnit
    ) -> RasterizedPattern:

        # dim_shape = DimShape(dim_shape)

        if not isinstance(dim_shape.shape, Spot):
            raise RuntimeError('Only `shapes.Spot`s can have `SpotStyle` as raster style.')

        spot = dim_shape.shape
        dwell_point = [
            *(spot.position * scale_factor(out_length_unit, dim_shape.unit)),
            scale_to(out_time_unit, mill.dwell_time)
        ]

        # return RasterizedPoints(np.array([dwell_point]*mill.repeats), False)
        return RasterizedPattern(np.array([dwell_point]*mill.repeats), out_length_unit, out_time_unit)



from typing import Union, Callable, Optional, Tuple, List

from fibomat.shapes import DimShape
from fibomat.linalg import Transformable, DimTransformable, Vector, DimVector, VectorLike, DimVectorLike
from fibomat.units import U_, scale_factor, LengthQuantity, has_length_dim
from fibomat.layout.lattices.lattice_base import LatticeBaseMixin
from fibomat.layout.groups.dim_group import DimGroup
from fibomat.shapes import Rect


class DimLattice(DimGroup, LatticeBaseMixin):
    def __init__(self, elements: List[DimTransformable], description: Optional[str] = None):
        super().__init__(elements, description=description)

    @classmethod
    def generate_rect(
        cls,
        nu: int, nv: int,
        dim_du: LengthQuantity, dim_dv: LengthQuantity,
        dim_element: Union[Transformable, Callable],
        dim_center: Optional[DimVectorLike] = None,
        predicate: Optional[Union[Callable, List[Callable]]] = None,
        explode: bool = False,
        remove_outliers: bool = False,
    ):
        nu = int(nu)
        nv = int(nv)

        if nu < 1 or nv < 1:
            raise ValueError('nu and nv must be at least 1.')

        if not has_length_dim(dim_du) or not has_length_dim(dim_dv):
            raise ValueError('dim_du and dim_dv must be length quantities')

        unit = dim_du.u
        du = dim_du.m
        dv = dim_dv.m_as(unit)

        seed_u = du / 2 if nu % 2 == 0 else 0
        seed_v = dv / 2 if nv % 2 == 0 else 0

        seed = Vector(seed_u, seed_v)

        # TODO: seed!
        return cls.generate(
            Rect(width=du*nu, height=dv*nv) * unit, (du, 0) * unit, (0, -dv) * unit, dim_element, dim_center, predicate, explode, remove_outliers, dim_seed=seed * unit
        )

    @classmethod
    def generate(
        cls,
        dim_boundary: DimShape,
        dim_u: VectorLike, dim_v: VectorLike,
        dim_element: Union[Transformable, Callable[[Tuple[float, float], Tuple[int, int]], Optional[DimTransformable]]],
        dim_center: Optional[VectorLike] = None,
        predicate: Optional[Union[Callable, List[Callable]]] = None,
        explode: bool = False,
        remove_outliers: bool = False,
        # break_layouts: bool = False
        dim_seed: Optional[DimVectorLike] = None
    ):
        """
        predicate get its coordinates in µm

        Args:
            dim_boundary:
            dim_u:
            dim_v:
            dim_element:
            dim_center:
            predicate:
            explode:
            remove_outliers:

        Returns:

        """
        base_unit = U_('µm')

        def dim_vec_to_vec(dim_vec: DimVector) -> Vector:
            return scale_factor(base_unit, dim_vec.unit) * dim_vec.vector

        def vec_to_dim_vec(vec: Vector) -> DimVector:
            return vec * base_unit

        u = dim_vec_to_vec(DimVector(dim_u))
        v = dim_vec_to_vec(DimVector(dim_v))

        if dim_seed is None:
            dim_seed = dim_center
        else:
            dim_seed = DimVector(dim_seed)

        seed = dim_vec_to_vec(dim_seed)

        center = dim_vec_to_vec(DimVector(dim_center))

        boundary = dim_boundary.scaled(scale_factor(base_unit, dim_boundary.unit), 'pivot').shape

        if callable(dim_element):
            element_gen = dim_element
        else:
            def element_gen(*args):
                return dim_element

        # TODO: seed!
        elements = cls._generate_impl(
            boundary, u, v, center, element_gen, predicate, explode, remove_outliers, dim_vec_to_vec, vec_to_dim_vec, seed
        )

        return cls(elements)



from __future__ import annotations
from typing import Optional, List, Any

from fibomat.layout.groups.group_base import GroupBase
from fibomat.linalg import DimVector, DimTransformable, DimBoundingBox


class DimGroup(GroupBase[DimTransformable, DimVector, DimBoundingBox], DimTransformable):
    def __init__(
        self,
        elements: List[DimTransformable],
        description: Optional[str] = None
    ):
        super().__init__(elements=elements, description=description)



from typing import List, Union, TYPE_CHECKING

import numpy as np

from fibomat.shapes.line import Line
from fibomat.shapes.arc import Arc
from fibomat.shapes.biarc import Biarc


if TYPE_CHECKING:
    from fibomat.shapes.parametric_curve import ParametricCurve

from fibomat.curve_tools.biarc_approximation.monotone_intervals import intervals_by_osculating_circles, IntervalType
from fibomat.curve_tools.biarc_approximation.spiral_segment import _approximate_param_curve_greedy
from fibomat.curve_tools.biarc_approximation.utils import _make_fixed_tangent_curvature


def approximate_parametric_curve(
    param_curve: 'ParametricCurve', rasterize_pitch: float, epsilon: float
) -> List[Union[Arc, Line]]:
    """Approximate a ParametricCurve with an ArcSpline.

    Seems to work good: rasterize_pitch > 10 * epsilon

    rasterize_pitch > epsilon

    Args:
        param_curve (ParametricCurve): curve to be approximated
        rasterize_pitch (float): pitch used for rasterization.
        epsilon (float): maximal distance between original and approximated curve.

    Returns:
        ArcSpline

    References:
        - https://www.sciencedirect.com/science/article/pii/037704279400029Z
        - https://epub.jku.at/obvulihs/content/titleinfo/2474921
        - https://www.sciencedirect.com/science/article/abs/pii/S0010448508001681
        - (https://www.sciencedirect.com/science/article/pii/S0925772112000326?via%3Dihub)

    """

    if rasterize_pitch <= epsilon:
        raise ValueError('rasterize_pitch must be large than epsilon.')

    parameter_values = param_curve.rasterize(rasterize_pitch, safety=1, add_endpoint=True)

    # For testing of FLUSH_WITHOUT_TRANSITION
    # parameter_values = np.roll(parameter_values, 10)

    monotone_curvature_intervals = intervals_by_osculating_circles(param_curve, parameter_values, epsilon)

    curve_segments = []

    for interval_type, monotone_interval in monotone_curvature_intervals:
        p_0 = param_curve.f(monotone_interval[0])
        p_1 = param_curve.f(monotone_interval[-1])

        t_0, _ = _make_fixed_tangent_curvature(param_curve, monotone_interval[0], epsilon, 'left')
        t_1, _ = _make_fixed_tangent_curvature(param_curve, monotone_interval[-1], epsilon, 'right')

        if interval_type == IntervalType.LINE:
            curve_segments.append(Line(p_0, p_1))
        elif interval_type == IntervalType.BIARC:
            # todo fix tangents
            curve_segments.append(Biarc(p_0, p_1, t_0, t_1))
        elif interval_type == IntervalType.ARC:
            if np.allclose(p_0, p_1):
                raise NotImplementedError(
                    'Cannot handle closed circles in biarc fitting currently. '
                    'Please report this as a bug if you need this.'
                )
            intermediate = param_curve.f(monotone_interval[len(monotone_interval) // 2])
            curve_segments.append(Arc.from_points(p_0, intermediate, p_1))
        else:
            new_segs = _approximate_param_curve_greedy(
                param_curve, monotone_interval, epsilon
            )
            curve_segments.extend(
                new_segs
            )

    # return parameter_values, curve_segments
    return curve_segments



"""Provide the :class:`Polyline` class."""
from __future__ import annotations
from typing import Optional, List

import numpy as np

from fibomat.shapes import shape
from fibomat.shapes.arc_spline import ArcSpline, ArcSplineCompatible
from fibomat.linalg import VectorLike, Vector, BoundingBox


class Polyline(shape.Shape, ArcSplineCompatible):
    """Polyline"""

    def __init__(self, points: List[VectorLike], description: Optional[str] = None, _closed: bool = False):
        """
        Args:
            points (VectorArrayLike): polyline points
            description (str, optional): description
        """
        super().__init__(description)

        # TODO: error checking?
        self._polygon = ArcSpline([(*vec, 0.) for vec in points], _closed)

    def to_arc_spline(self) -> ArcSpline:
        return self._polygon.clone_with_new_description(self.description)

    @property
    def points(self) -> np.ndarray:
        """Polyline points.

        Access:
            get

        Returns:
            np.ndarray
        """
        return self._polygon.vertices[:, :2]

    @property
    def n_points(self) -> int:
        """Number of points.

        Access:
            get

        Returns:
            int
        """
        return len(self._polygon)

    # def clone(self) -> Polyline:
    #     return self.__class__(self._points.clone())
    #
    # __copy__ = clone

    def __repr__(self) -> str:
        if self.n_points < 5:
            return '{}(points={!r})'.format(
                self.__class__.__name__, self.points)
        else:
            return '{}(points=...)'.format(
                self.__class__.__name__)

    @property
    def bounding_box(self) -> BoundingBox:
        return self._polygon.bounding_box

    @property
    def is_closed(self) -> bool:
        return self._polygon.is_closed

    @property
    def center(self) -> Vector:
        return self._polygon.center

    def _impl_translate(self, trans_vec: VectorLike) -> None:
        self._polygon._impl_translate(trans_vec)  # pylint: disable=protected-access

    def _impl_rotate(self, theta: float) -> None:
        self._polygon._impl_rotate(theta)  # pylint: disable=protected-access

    def _impl_scale(self, fac: float) -> None:
        self._polygon._impl_scale(fac)  # pylint: disable=protected-access

    def _impl_mirror(self, mirror_axis: VectorLike) -> None:
        self._polygon._impl_mirror(mirror_axis)  # pylint: disable=protected-access



from typing import Callable

import numpy as np

from vispy import app
from vispy.visuals.transforms import STTransform

from fibomat.beam_simulation.markers import Tail, DwellPoints
from fibomat.beam_simulation.scalebar import ScaleBar


class FIBCanvas(app.Canvas):

    def __init__(self, dwell_points: np.ndarray, length_unit: str, on_timer_callback: Callable):
        app.Canvas.__init__(self, keys='interactive', size=(800, 600))

        dwell_points = np.asarray(dwell_points, dtype=np.float32)

        min_x, min_y = np.min(dwell_points, axis=0)
        max_x, max_y = np.max(dwell_points, axis=0)

        if self.physical_size[0] / (max_x - min_x) < self.physical_size[1] / (max_y - min_y):
            self._pre_scale = self.physical_size[0] / (max_x - min_x)
        else:
            self._pre_scale = self.physical_size[1] / (max_y - min_y)

        dwell_points *= self._pre_scale

        self._min_x, self._min_y = np.min(dwell_points, axis=0)
        self._max_x, self._max_y = np.max(dwell_points, axis=0)

        self._n_points = len(dwell_points)

        self._tail = Tail(dwell_points, 100, 10)
        self._dwell_points = DwellPoints(dwell_points)
        self._scale_bar = ScaleBar(length_unit, self)

        w, h = self.size
        self._dwell_points.transform = STTransform(translate=(w / 2., h / 2.))
        self._tail.transform = STTransform(translate=(w / 2., h / 2.))

        self._timer = app.Timer('auto', connect=self.on_timer, start=True)
        self._current_point = 0
        self._last_update = 0
        self._speed = 10
        self._run_animation = False

        self._on_timer_callback = on_timer_callback

    def on_resize(self, event):
        vp = (0, 0, self.physical_size[0], self.physical_size[1])
        self.context.set_viewport(*vp)

        self._dwell_points.transforms.configure(canvas=self, viewport=vp)
        self._dwell_points.transform.translate = self.physical_size[0] / 2, self.physical_size[1] / 2

        self._tail.transforms.configure(canvas=self, viewport=vp)
        self._tail.transform.translate = self.physical_size[0] / 2, self.physical_size[1] / 2

        width = self._max_x - self._min_x
        height = self._max_y - self._min_y

        margin = 0  # 10

        if self.physical_size[0] / (self._max_x - self._min_x) < self.physical_size[1] / (self._max_y - self._min_y):
            s = self.physical_size[0] / (width+margin)
            pixel_per_length = (self._max_x - self._min_x) / (self.physical_size[0] - 2*margin)
        else:
            s = self.physical_size[1] / (height+margin)
            pixel_per_length = (self._max_y - self._min_y) / (self.physical_size[1] - 2*margin)

        self._dwell_points.transform.scale = s, s
        self._tail.transform.scale = s, s
        self._scale_bar.transform(pixel_per_length / self._pre_scale, s)

    def on_draw(self, event):
        self.context.clear('white')
        self._dwell_points.draw()
        self._tail.draw(self._current_point)
        self._scale_bar.draw()

    def on_timer(self, event):
        if self._run_animation:
            freq = 1 / self._speed

            elapsed = self._timer.elapsed
            delta = elapsed - self._last_update

            if delta > freq:
                n_new_points = int(delta / freq)
                self._current_point += n_new_points
                self._last_update = elapsed
                self._on_timer_callback(self._current_point)
        self.update()

    def start_animation(self):
        # self.timer.start()
        self._run_animation = True
        self._last_update = self._timer.elapsed

    def stop_animation(self):
        # self.timer.stop()
        self._run_animation = False

    def reset_animation(self):
        self._last_update = 0
        self._current_point = 0
        self._run_animation = False

    def set_speed(self, speed):
        self._speed = speed

    def set_tail_length(self, length):
        self._tail._tail_length = length

    def set_alpha(self, alpha):
        self._tail._alpha = alpha

    @property
    def number_of_points(self) -> int:
        return self._n_points

    @property
    def current_point_index(self) -> int:
        return self._current_point

    @current_point_index.setter
    def current_point_index(self, value):
        self._current_point = value
